import os

import numpy as np
import tensorflow as tf
from tensorflow.python.framework import test_util

from dynastes.layers.spectral_layers import Wave2STFTLayer, STFT2MelspectrogramLayer
from dynastes.util.test_utils import layer_test


class SpectralLayersTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        print('### Running spectral ops tests ###')
        for fft_bins in [128, 256, 512]:
            for hop_div in [2, 4, 8]:
                hop_length = fft_bins // hop_div
                for pad_mul in [2, 3, 4, 7]:
                    pad = fft_bins * pad_mul
                    n_mels = 128
                    dir_path = os.path.dirname(os.path.realpath(__file__))

                    raw_audio = tf.io.read_file(dir_path + '/data/test_wav.wav')
                    waves = tf.audio.decode_wav(raw_audio)[0][:1024] * 0.25
                    waves = tf.expand_dims(waves, 0)
                    waves = tf.expand_dims(waves, -1)
                    waves = tf.keras.layers.UpSampling2D((2, 1), interpolation='bilinear')(waves)
                    waves: tf.Tensor = tf.squeeze(waves, axis=-1)

                    wav2stft_layer = Wave2STFTLayer(n_bins=fft_bins,
                                                    hop_length=hop_length,
                                                    padding=pad)

                    stft2mel_layer = STFT2MelspectrogramLayer(n_mels=n_mels,
                                                              min_hz=40.,
                                                              max_hz=16000.,
                                                              sr=44100,
                                                              ifreq=True,
                                                              return_phase=True,
                                                              trainable=True)
                    with tf.GradientTape() as t:

                        stfts = wav2stft_layer(waves)

                        comp_stfts_out_shape = wav2stft_layer.compute_output_shape(waves.shape.as_list())
                        stfts_shape = stfts.shape.as_list()
                        try:
                            self.assertEqual(stfts_shape, comp_stfts_out_shape)
                        except AssertionError as err:
                            print(fft_bins, hop_length, pad)
                            raise err

                        melspecs = stft2mel_layer(stfts)
                        comp_melspecs_out_shape = stft2mel_layer.compute_output_shape(stfts_shape)
                        melspecs_shape = melspecs.shape.as_list()
                        self.assertEqual(comp_melspecs_out_shape, melspecs_shape)

                        istfts = stft2mel_layer(melspecs, invert=True)
                        iwaves = wav2stft_layer(istfts, invert=True)

                        loss = tf.keras.losses.mean_absolute_error(waves, iwaves)

                    grads = t.gradient(loss, stft2mel_layer.trainable_weights)
                    for grad in grads:
                        self.assertNotAllClose(grad, np.zeros_like(grad))


class Wav2STFTLayerTest(tf.test.TestCase):

    def test_simple(self):
        layer_test(
            Wave2STFTLayer, kwargs={'n_bins': 128, 'hop_length': 64, 'padding': 64, 'hq': True},
            input_shape=(5, 32, 2),
            expected_output_dtype=tf.complex128)


class STFT2MelSpecLayerTest(tf.test.TestCase):

    def test_simple(self):
        layer_test(
            STFT2MelspectrogramLayer, kwargs={'n_mels': 128},
            input_shape=(5, 32, 128, 2),
            input_dtype=tf.complex128,
            expected_output_dtype=tf.float32)



import tensorflow as tf

from dynastes.ops.t2t_common import shape_list


def crop(x, crops):
    crops = tf.convert_to_tensor(crops)
    begins, ends = tf.unstack(crops, axis=-1)
    shape = tf.convert_to_tensor(shape_list(x))
    return tf.slice(x, begins, shape - (ends + begins))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from functools import partial

import tensorflow as tf
import tensorflow.keras.layers as tfkl

from dynastes import activations
from dynastes.blocks import layer_factory
from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.util import cache_context
from dynastes.util.layer_util import call_masked as cm


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class PointWiseFeedForwardBlock(DynastesBaseLayer):

    def __init__(self,
                 dff,
                 d_model,
                 kernel_size=1,
                 ff_type='Dense',
                 d_type='Dense',
                 depth_multiplier=1,
                 strides=1,
                 dilation_rate=1,
                 grouped=False,
                 group_size=1,
                 padding='same',
                 activation=None,
                 use_bias=False,
                 dropout_rate=0.,
                 **kwargs):
        super(PointWiseFeedForwardBlock, self).__init__(**kwargs)
        self.dff = dff
        self.d_model = d_model
        self.ff_type = ff_type
        self.d_type = d_type
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.supports_masking = True

        self.kernel_size = kernel_size
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.depth_multiplier = depth_multiplier
        self.padding = padding
        self.grouped = grouped
        self.group_size = group_size
        self.dropout_rate = dropout_rate

        conv_partial = partial(layer_factory.get_1d_layer, kernel_size=kernel_size,
                               grouped=grouped,
                               group_size=group_size,
                               depth_multiplier=depth_multiplier,
                               padding=padding,
                               use_bias=use_bias,
                               strides=strides,
                               dilation_rate=dilation_rate,
                               kernel_initializer=self.get_initializer('kernel'),
                               bias_initializer=self.get_initializer('bias'),
                               kernel_regularizer=self.get_regularizer('kernel'),
                               bias_regularizer=self.get_regularizer('bias'),
                               activity_regularizer=None,
                               kernel_constraint=self.get_constraint('kernel'),
                               bias_constraint=self.get_constraint('bias'))

        self.dff_layer = conv_partial(
            type=ff_type,
            filters=dff,
            activation=self.activation)
        self.out_layer = conv_partial(
            type=d_type,
            filters=d_model,
            activation=None)

    def call_masked(self, inputs, training=None, mask=None):
        x, x_mask = cm(self.dff_layer, inputs, training=training, mask=mask)
        x, x_mask = cm(self.out_layer, x, training=training, mask=x_mask)
        return x, x_mask

    def call(self, inputs, training=None, mask=None):
        x, x_mask = cm(self.dff_layer, inputs, training=training, mask=mask)
        x = self.out_layer(x, training=training, mask=x_mask)
        return x

    def compute_mask(self, inputs, mask=None):
        if mask is not None:
            if self.dff_layer.supports_masking:
                mask = self.dff_layer.compute_mask(inputs, mask=mask)
            if self.out_layer.supports_masking:
                mask = self.out_layer.compute_mask(inputs, mask=mask)
        return mask

    def compute_output_shape(self, input_shape):
        shape = self.dff_layer.compute_output_shape(input_shape)
        shape = self.out_layer.compute_output_shape(shape)
        return shape

    def get_config(self):
        config = {
            'dff': self.dff,
            'd_model': self.d_model,
            'ff_type': self.ff_type,
            'd_type': self.d_type,
            'activation': activations.serialize(self.activation),
            'use_bias': self.use_bias,
            'depth_multiplier': self.depth_multiplier,
            'kernel_size': self.kernel_size,
            'strides': self.strides,
            'dilation_rate': self.dilation_rate,
            'padding': self.padding,
            'grouped': self.grouped,
            'group_size': self.group_size,
            'dropout_rate': self.dropout_rate,
        }
        base_config = super(PointWiseFeedForwardBlock, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class EncoderBlock(tfkl.Layer):

    def __init__(self,
                 sa_layer: tfkl.Layer,
                 norm0: tfkl.Layer,
                 ffn: tfkl.Layer,
                 norm1: tfkl.Layer,
                 mha_skip_adapt: tfkl.Layer = tfkl.Activation('linear'),
                 ffn_skip_adapt: tfkl.Layer = tfkl.Activation('linear'),
                 **kwargs):
        super(EncoderBlock, self).__init__(**kwargs)
        self.sa_layer = sa_layer
        self.norm0 = norm0
        self.ffn = ffn
        self.norm1 = norm1
        self.mha_skip_adapt = mha_skip_adapt
        self.ffn_skip_adapt = ffn_skip_adapt

    def request_cache(self, batch_size=1, max_length=1):
        try:
            return self.sa_layer.request_cache(batch_size=batch_size, max_length=max_length)
        except:
            return None

    def call_masked(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None):
        with cache_context.SubContext(self.name):
            x = inputs
            x, x_mask = cm(self.sa_layer, x, training=training, mask=mask, cache=cache,
                           decode_loop_step=decode_loop_step)
            res, res_mask = cm(self.mha_skip_adapt, inputs, training=training, mask=mask)
            if x_mask is None:
                n_mask = res_mask
            elif res_mask is None:
                n_mask = x_mask
            else:
                n_mask = tf.math.logical_and(x_mask, res_mask)
            x, x_mask = cm(self.norm0, x + res, training=training, mask=n_mask)

            f, f_mask = cm(self.ffn, x, training=training, mask=x_mask)
            res, res_mask = cm(self.ffn_skip_adapt, x, training=training, mask=x_mask)
            if f_mask is None:
                n_mask = f_mask
            elif res_mask is None:
                n_mask = x_mask
            else:
                n_mask = tf.math.logical_and(f_mask, res_mask)
            x, mask = cm(self.norm1, f + res, training=training, mask=n_mask)
            return x, mask

    def call(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None):
        with cache_context.SubContext(self.name):
            x = inputs
            x, x_mask = cm(self.sa_layer, x, training=training, mask=mask, cache=cache,
                           decode_loop_step=decode_loop_step)
            res, res_mask = cm(self.mha_skip_adapt, inputs, training=training, mask=mask)
            if x_mask is None:
                n_mask = res_mask
            elif res_mask is None:
                n_mask = x_mask
            else:
                n_mask = tf.math.logical_and(x_mask, res_mask)
            x, x_mask = cm(self.norm0, x + res, training=training, mask=n_mask)

            f, f_mask = cm(self.ffn, x, training=training, mask=x_mask)
            res, res_mask = cm(self.ffn_skip_adapt, x, training=training, mask=x_mask)
            if f_mask is None:
                n_mask = f_mask
            elif res_mask is None:
                n_mask = x_mask
            else:
                n_mask = tf.math.logical_and(f_mask, res_mask)
            x, _ = cm(self.norm1, f + res, training=training, mask=n_mask)
            return x

    def compute_mask(self, inputs, mask=None):
        with cache_context.SubContext(self.name):
            x = inputs
            if self.sa_layer.supports_masking:
                x_mask = self.sa_layer.compute_mask(x, mask=mask)
            else:
                x_mask = mask
            if self.mha_skip_adapt.supports_masking:
                res_mask = self.mha_skip_adapt.compute_mask(inputs, mask=mask)
            else:
                res_mask = mask
            if x_mask is None:
                n_mask = res_mask
            elif res_mask is None:
                n_mask = x_mask
            else:
                n_mask = tf.math.logical_and(x_mask, res_mask)
            if self.norm0.supports_masking:
                x_mask = self.norm0.compute_mask(x, mask=n_mask)
            else:
                x_mask = n_mask
            if self.ffn.supports_masking:
                f_mask = self.ffn.compute_mask(x, mask=x_mask)
            else:
                f_mask = x_mask
            if self.ffn_skip_adapt.supports_masking:
                res_mask = self.ffn_skip_adapt.compute_mask(x, mask=x_mask)
            else:
                res_mask = x_mask
            if f_mask is None:
                n_mask = f_mask
            elif res_mask is None:
                n_mask = x_mask
            else:
                n_mask = tf.math.logical_and(f_mask, res_mask)
            if self.norm1.supports_masking:
                mask = self.norm1.compute_mask(x, mask=n_mask)
            else:
                mask = n_mask
            return mask

    def compute_output_shape(self, input_shape):
        s = input_shape
        s = self.mha_skip_adapt.compute_output_shape(s)
        s = self.ffn_skip_adapt.compute_output_shape(s)
        return s


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class EncoderBlockStack(tfkl.Layer):

    def __init__(self,
                 blocks,
                 **kwargs):
        super(EncoderBlockStack, self).__init__(**kwargs)
        self.blocks = blocks

    def request_cache(self, batch_size=1, max_length=1):
        cache = {}
        for i, block in enumerate(self.blocks):
            try:
                block_cache = block.request_cache(batch_size=batch_size, max_length=max_length)
            except:
                block_cache = None
            cache[i] = block_cache
        return cache

    def call_masked(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None, **kwargs):
        with cache_context.SubContext(self.name):
            x = inputs
            for i, block in enumerate(self.blocks):
                if cache is not None:
                    block_cache = cache[i]
                else:
                    block_cache = None
                x, mask = cm(block, x, training=training, mask=mask, cache=block_cache,
                             decode_loop_step=decode_loop_step,
                             **kwargs)
            return x, mask

    def call(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None, **kwargs):
        with cache_context.SubContext(self.name):
            x = inputs
            for i, block in enumerate(self.blocks):
                if cache is not None:
                    block_cache = cache[i]
                else:
                    block_cache = None
                x = block(x, training=training, mask=mask, cache=block_cache, decode_loop_step=decode_loop_step,
                          **kwargs)
            return x

    def compute_mask(self, inputs, mask=None):
        with cache_context.SubContext(self.name):
            x = inputs
            for block in self.blocks:
                if block.supports_masking:
                    mask = block.compute_mask(x, mask=mask)
            return mask

    def compute_output_shape(self, input_shape):
        s = input_shape
        for block in self.blocks:
            s = block.compute_output_shape(s)
        return s


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class DecoderBlock(tfkl.Layer):

    def __init__(self,
                 sa_layer: tfkl.Layer,
                 norm0: tfkl.Layer,
                 ca_layer: tfkl.Layer,
                 norm1: tfkl.Layer,
                 ffn: tfkl.Layer,
                 norm2: tfkl.Layer,
                 mha_skip_adapt: tfkl.Layer = tfkl.Activation('linear'),
                 ffn_skip_adapt: tfkl.Layer = tfkl.Activation('linear'),
                 **kwargs):
        super(DecoderBlock, self).__init__(**kwargs)
        self.sa_layer = sa_layer
        self.norm0 = norm0
        self.ca_layer = ca_layer
        self.norm1 = norm1
        self.ffn = ffn
        self.norm2 = norm2
        self.mha_skip_adapt = mha_skip_adapt
        self.ffn_skip_adapt = ffn_skip_adapt

    def request_cache(self, batch_size=1, max_length_sa=1, max_length_ca=1):
        try:
            return {'sa': self.sa_layer.request_cache(batch_size=batch_size, max_length=max_length_sa),
                    'ca': self.ca_layer.request_cache(batch_size=batch_size, max_length=max_length_ca)}
        except:
            return None

    def _call(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None, pad_q_to_kv=False):
        x, enc_in = inputs
        _x = x
        if mask is not None:
            x_mask, enc_mask = mask
        else:
            assert False
            x_mask = None
            enc_mask = None
        _x_mask = x_mask
        ## Self-attention
        if cache is not None:
            sa_cache = cache['sa']
        else:
            sa_cache = None
        x, x_mask = cm(self.sa_layer, x, training=training, mask=x_mask, cache=sa_cache,
                       decode_loop_step=decode_loop_step, pad_q_to_kv=pad_q_to_kv)
        res, res_mask = cm(self.mha_skip_adapt, _x, training=training, mask=_x_mask)
        if x_mask is None:
            n_mask = res_mask
        elif res_mask is None:
            n_mask = x_mask
        else:
            n_mask = tf.math.logical_and(x_mask, res_mask)
        x, x_mask = cm(self.norm0, x + res, training=training, mask=n_mask)
        _x = x
        ## Attend to encoding
        if mask is not None:
            ca_mask = x_mask, enc_mask
        else:
            assert False
            ca_mask = None
        if cache is not None:
            ca_cache = cache['ca']
        else:
            ca_cache = None
        x, x_mask = cm(self.ca_layer, (x, enc_in), training=training, mask=ca_mask, cache=ca_cache)

        res, res_mask = cm(self.mha_skip_adapt, _x, training=training, mask=x_mask)
        if x_mask is None:
            n_mask = res_mask
        elif res_mask is None:
            n_mask = x_mask
        else:
            n_mask = tf.math.logical_and(x_mask, res_mask)

        x, x_mask = cm(self.norm1, x + res, training=training, mask=n_mask)
        _x = x
        ## FF-net
        f, f_mask = cm(self.ffn, x, training=training, mask=x_mask)
        res, res_mask = cm(self.ffn_skip_adapt, _x, training=training, mask=x_mask)
        if f_mask is None:
            n_mask = f_mask
        elif res_mask is None:
            n_mask = x_mask
        else:
            n_mask = tf.math.logical_and(f_mask, res_mask)
        x, mask = cm(self.norm2, f + res, training=training, mask=n_mask)

        return x, mask

    def call_masked(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None, pad_q_to_kv=False):
        return self._call(inputs, training=training, mask=mask, cache=cache, decode_loop_step=decode_loop_step, pad_q_to_kv=pad_q_to_kv)

    def call(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None, pad_q_to_kv=False):
        x, _ = self._call(inputs, training=training, mask=mask, cache=cache, decode_loop_step=decode_loop_step, pad_q_to_kv=pad_q_to_kv)
        return x

    def compute_mask(self, inputs, mask=None):
        return mask

    def compute_output_shape(self, input_shape):
        s = input_shape
        s = self.mha_skip_adapt.compute_output_shape(s)
        s = self.ffn_skip_adapt.compute_output_shape(s)
        return s


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class DecoderBlockStack(tfkl.Layer):

    def __init__(self,
                 blocks,
                 **kwargs):
        super(DecoderBlockStack, self).__init__(**kwargs)
        self.blocks = blocks

    def request_cache(self, batch_size=1, max_length_ca=1, max_length_sa=1):
        cache = {}
        for i, block in enumerate(self.blocks):
            try:
                block_cache = block.request_cache(batch_size=batch_size, max_length_ca=max_length_ca,
                                                  max_length_sa=max_length_sa)
            except:
                block_cache = None
            cache[i] = block_cache
        return cache

    def call_masked(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None, **kwargs):
        x = inputs
        for i, block in enumerate(self.blocks):
            if cache is not None:
                block_cache = cache[i]
            else:
                block_cache = None
            x, mask = cm(block, x, training=training, mask=mask, cache=block_cache, decode_loop_step=decode_loop_step,
                         **kwargs)
        return x, mask

    def call(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None, pad_q_to_kv=False, **kwargs):
        x, enc = inputs
        for i, block in enumerate(self.blocks):
            if cache is not None:
                block_cache = cache[i]
            else:
                block_cache = None
            x = block((x, enc), training=training, mask=mask, cache=block_cache, decode_loop_step=decode_loop_step,
                      pad_q_to_kv=pad_q_to_kv,
                      **kwargs)
        return x

    def compute_mask(self, inputs, mask=None):
        x = inputs
        for block in self.blocks:
            if block.supports_masking:
                mask = block.compute_mask(x, mask=mask)
        return mask

    def compute_output_shape(self, input_shape):
        s = input_shape
        for block in self.blocks:
            s = block.compute_output_shape(s)
        return s



import tensorflow as tf

from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.ops import t2t_attention
from dynastes.ops.t2t_common import shape_list
from dynastes.probability.pseudoblocksparse_bijectors import PseudoBlockSparseBijector1D
from dynastes.util.precision_util import large_compatible_negative


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class AddTimingSignalLayer1D(DynastesBaseLayer):

    def __init__(self,
                 min_timescale=1.0,
                 max_timescale=1.0e4,
                 **kwargs):
        super(AddTimingSignalLayer1D, self).__init__(**kwargs)
        self.min_timescale = min_timescale
        self.max_timescale = max_timescale

    def call(self, inputs, **kwargs):
        return t2t_attention.add_timing_signal_1d(inputs, min_timescale=self.min_timescale,
                                                  max_timescale=self.max_timescale)

    def get_config(self):
        config = {'min_timescale': self.min_timescale,
                  'max_timescale': self.max_timescale, }
        base_config = super(AddTimingSignalLayer1D, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class LshGatingLayer(DynastesBaseLayer):

    def __init__(self,
                 bucket_length,
                 trainable=False,
                 **kwargs):
        super(LshGatingLayer, self).__init__(trainable=trainable, **kwargs)
        self.bucket_length = bucket_length
        self.buckets = 2 ** bucket_length

    def _idx_to_bits(self, i):
        """Convert an group index to its bit representation."""
        bits = bin(i)[2:].zfill(self.bucket_length)  # Pad the bits str with 0
        return [-1.0 if b == "0" else 1.0 for b in bits]

    def build(self, input_shape):
        self.t_vectors = self.add_weight(name='vector', shape=(input_shape[-1], self.bucket_length), dtype=self.dtype,
                                         initializer=tf.keras.initializers.RandomNormal(),
                                         trainable=self.trainable)
        self.t_group = tf.constant(
            [self._idx_to_bits(i) for i in range(self.buckets)],
            dtype=self.dtype,
            name="group")

    def call(self, inputs, training=None):
        # The balance loss don't propagate to the rest of the network
        x = tf.stop_gradient(inputs)
        # [length, depth] * [depth, nb_vectors * replicat]
        x = tf.matmul(x, self.get_weight('vector', training=training))
        # [length, nb_vector * replicat]
        x = tf.sign(x)  # Get on which side of the hyperplane the keys are.

        # x = tf.reshape(x, [-1, nb_replicat, nb_vector])
        # [length, replicat, nb_vector] * [nb_vector, 2^nb_vector - 1]

        x = tf.matmul(x, self.t_group, transpose_b=True) / self.bucket_length
        # We get a similarity score for each of the group between [-1, 1]
        # [length, (replicat,) 2^nb_vector - 1]
        # Do an argmax to get the most likely group for each replicat
        x = tf.argmax(x, axis=-1)
        # [length(, replicat)]
        # One-hot for compatibility with the sparse dispatcher
        x = tf.one_hot(x, self.buckets + 1)
        # TODO(epot): Use a loss to force an even distribution
        return x

    def get_config(self):
        config = {'bucket_length': self.bucket_length}
        base_config = super(LshGatingLayer, self).get_config()
        return {**base_config, **config}


def _get_attention_type_1D(local, masked, relative, self_attention, sparse):
    if self_attention:
        if sparse:
            if local:
                raise NotImplementedError
            elif relative:
                raise NotImplementedError
            else:
                attention_type = 'sparse_attention_truncated'

        else:
            if local:
                if masked:
                    attention_type = 'masked_local_attention_1d'
                else:
                    attention_type = 'unmasked_local_attention_1d'
            else:
                if relative:
                    if masked:
                        attention_type = 'masked_self_attention_relative'
                    else:
                        attention_type = 'unmasked_self_attention_relative'
                else:
                    attention_type = 'attention'
    else:
        if sparse:
            raise NotImplementedError
        if local:
            if masked:
                attention_type = 'masked_local_attention_1d'
            else:
                attention_type = 'unmasked_local_attention_1d'
        else:
            if relative:
                raise NotImplementedError
            else:
                attention_type = 'attention'
    return attention_type


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class Attention1D(DynastesBaseLayer):

    def __init__(self,
                 num_heads,
                 multiquery_attention=False,
                 self_attention=False,
                 masked=False,
                 local=False,
                 relative=False,
                 sparse=False,
                 dropout_rate=0.0,
                 max_relative_position=2,
                 lsh_bucket_length=4,
                 block_length=128,
                 filter_width=100,
                 mask_right=False,
                 add_relative_to_values=False,
                 heads_share_relative_embeddings=False,
                 **kwargs):
        super(Attention1D, self).__init__(**kwargs)

        attention_type = _get_attention_type_1D(local, masked, relative, self_attention, sparse)

        assert (attention_type in [
            'attention',
            'unmasked_self_attention_relative',
            'masked_self_attention_relative',
            'unmasked_local_attention_1d',
            'masked_local_attention_1d',
            'sparse_attention_truncated'])
        self.num_heads = num_heads
        self.multiquery_attention = multiquery_attention
        if self.multiquery_attention:
            self.num_heads_kv = 1
        else:
            self.num_heads_kv = num_heads
        self.attention_type = attention_type
        self.masked = masked
        self.local = local
        self.relative = relative
        self.sparse = sparse
        self.mask_right = mask_right
        self.dropout_rate = dropout_rate
        self.lsh_bucket_length = lsh_bucket_length
        self.max_relative_position = max_relative_position
        self.block_length = block_length
        self.filter_width = filter_width
        self.lsh_gates = None
        self.add_relative_to_values = add_relative_to_values
        self.maybe_build_lsh_gates()
        self.heads_share_relative_embeddings = heads_share_relative_embeddings

    def maybe_build_lsh_gates(self):
        if self.attention_type == 'sparse_attention_truncated' and self.num_heads is not None and self.lsh_gates is None:
            self.lsh_gates = [LshGatingLayer(bucket_length=self.lsh_bucket_length) for _ in range(self.num_heads)]

    def build(self, input_shape):
        self.depth_q = int(input_shape[0][-1]) // self.num_heads
        depth_k = int(input_shape[1][-1]) // self.num_heads_kv
        depth_v = int(input_shape[2][-1]) // self.num_heads_kv
        if 'relative' in self.attention_type:
            if self.attention_type == 'unmasked_self_attention_relative':
                k_embedding_shape = t2t_attention.get_relative_embeddings_left_right_shape(
                    self.max_relative_position,
                    depth_k, self.num_heads,
                    self.heads_share_relative_embeddings)
                self.key_embeddings = self.add_weight('key_embeddings',
                                                      initializer=tf.keras.initializers.RandomNormal(
                                                          stddev=t2t_attention.get_embedding_initializer_stddev(
                                                              depth_k)),
                                                      shape=k_embedding_shape)

                if self.add_relative_to_values:
                    v_embedding_shape = t2t_attention.get_relative_embeddings_left_right_shape(
                        self.max_relative_position,
                        depth_v, self.num_heads,
                        False)
                    self.value_embeddings = self.add_weight('value_embeddings',
                                                            initializer=tf.keras.initializers.RandomNormal(
                                                                stddev=t2t_attention.get_embedding_initializer_stddev(
                                                                    depth_v)),
                                                            shape=v_embedding_shape)
                else:
                    self.value_embeddings = None
            elif self.attention_type == 'masked_self_attention_relative':
                k_embedding_shape = t2t_attention.get_relative_embeddings_left_shape(
                    self.max_relative_position,
                    depth_k, self.num_heads, False)
                self.key_embeddings = self.add_weight('key_embeddings',
                                                      initializer=tf.keras.initializers.RandomNormal(
                                                          stddev=t2t_attention.get_embedding_initializer_stddev(
                                                              depth_k)),
                                                      shape=k_embedding_shape)
                if self.add_relative_to_values:
                    v_embedding_shape = t2t_attention.get_relative_embeddings_left_shape(
                        self.max_relative_position,
                        depth_v, self.num_heads,
                        False)
                    self.value_embeddings = self.add_weight('value_embeddings',
                                                            initializer=tf.keras.initializers.RandomNormal(
                                                                stddev=t2t_attention.get_embedding_initializer_stddev(
                                                                    depth_v)),
                                                            shape=v_embedding_shape)
                else:
                    self.value_embeddings = None
            else:
                raise ValueError()

        self.maybe_build_lsh_gates()
        super(Attention1D, self).build(input_shape)

    def _create_look_ahead_mask(self, size):
        mask = 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0)
        return mask  # (seq_len, seq_len)

    def call(self, inputs, training=None, mask=None):
        if len(inputs) == 3:
            q, k, v = inputs
        else:
            raise ValueError()

        q_shape = shape_list(q)

        if mask is not None and self.attention_type != 'masked_local_attention_1d':
            mask_0_shape = shape_list(mask[0])
            mask_1_shape = shape_list(mask[1])
            q_mask = (1. - tf.cast(mask[0], tf.float32))[:, tf.newaxis, :, tf.newaxis]
            if self.mask_right and q_shape[1] is not None:
                # TODO: Reenable this somehow

                """
                @tf.function
                def assert_mask_ok(mask_0_shape, mask_1_shape):
                    assert mask_0_shape[1] == mask_1_shape[1] or mask_0_shape[1] == 1

                assert_mask_ok(mask_0_shape, mask_1_shape)
                """
                look_ahead_mask = self._create_look_ahead_mask(q_shape[1])
                q_mask = tf.maximum(q_mask, look_ahead_mask)
            kv_mask = (1. - tf.cast(mask[1], tf.float32))[:, tf.newaxis, tf.newaxis, :]
            c_mask = tf.maximum(q_mask, kv_mask)

            # c_mask = tf.maximum(c_mask, look_ahead_mask)
            bias = c_mask * large_compatible_negative(k.dtype)
        else:
            if self.attention_type != 'masked_local_attention_1d' and self.mask_right:
                look_ahead_mask = self._create_look_ahead_mask(q_shape[1])
                bias = look_ahead_mask * large_compatible_negative(k.dtype)
            else:
                bias = None

        r = None
        weights = None

        q = t2t_attention.split_heads(q, self.num_heads)
        k = t2t_attention.split_heads(k, self.num_heads_kv)
        v = t2t_attention.split_heads(v, self.num_heads_kv)
        if training:
            rate = self.dropout_rate
        else:
            rate = 0.

        if 'relative' in self.attention_type:
            key_embeddings = self.get_weight('key_embeddings', training=training)
            if self.add_relative_to_values:
                value_embeddings = self.get_weight('value_embeddings', training=training)
            else:
                value_embeddings = None
            if self.attention_type == 'unmasked_self_attention_relative':
                r, weights = t2t_attention.dot_product_unmasked_self_attention_relative_v2(q=q, k=k, v=v, bias=bias,
                                                                                           key_leftright_embeddings=key_embeddings,
                                                                                           value_leftright_embeddings=value_embeddings,
                                                                                           dropout_rate=rate,
                                                                                           max_relative_position=self.max_relative_position,
                                                                                           heads_share_relative_embedding=self.heads_share_relative_embeddings)
            elif self.attention_type == 'masked_self_attention_relative':
                r, weights = t2t_attention.dot_product_self_attention_relative_v2(q=q, k=k, v=v, bias=bias,
                                                                                  key_left_embedding=key_embeddings,
                                                                                  value_left_embedding=value_embeddings,
                                                                                  dropout_rate=rate,
                                                                                  max_relative_position=self.max_relative_position,
                                                                                  heads_share_relative_embedding=self.heads_share_relative_embeddings)
        else:
            if self.attention_type == 'unmasked_local_attention_1d':
                r, weights = t2t_attention.local_attention_1d(q=q, k=k, v=v, block_length=self.block_length,
                                                              filter_width=self.filter_width)
            elif self.attention_type == 'masked_local_attention_1d':
                if mask is not None:
                    attn_mask = tf.cast(mask[1], k.dtype)
                else:
                    attn_mask = None
                r, weights = t2t_attention.masked_local_attention_1d(q=q, k=k, v=v, block_length=self.block_length,
                                                                     mask_right=self.mask_right,
                                                                     mask=attn_mask,
                                                                     dropout_rate=rate)
            elif self.attention_type == 'sparse_attention_truncated':
                r, loss, weights = t2t_attention.sparse_dot_product_attention_truncated(q=q, k=k, v=v,
                                                                                        list_lsh=self.lsh_gates,
                                                                                        mask_right=self.mask_right)
                self.add_loss(loss)
            else:
                r, weights = t2t_attention.dot_product_attention(q=q, k=k, v=v, bias=bias,
                                                                 dropout_rate=rate)

        r = t2t_attention.combine_heads(r)
        return r, weights

    def compute_output_shape(self, input_shape):
        depth_v = int(input_shape[2][-1]) // self.num_heads_kv
        output_shape = input_shape[0][:-1] + [depth_v * self.num_heads]
        attn_w_shape = [input_shape[0][0], input_shape[0][1], input_shape[1][1]]
        return output_shape, attn_w_shape

    def get_config(self):
        config = {
            'num_heads': self.num_heads,
            'multiquery_attention': self.multiquery_attention,
            'masked': self.masked,
            'local': self.local,
            'relative': self.relative,
            'sparse': self.sparse,
            'dropout_rate': self.dropout_rate,
            'max_relative_position': self.max_relative_position,
            'lsh_bucket_length': self.lsh_bucket_length,
            'block_length': self.block_length,
            'filter_width': self.filter_width,
            'mask_right': self.mask_right,
            'add_relative_to_values': self.add_relative_to_values,
            'heads_share_relative_embeddings': self.heads_share_relative_embeddings
        }
        base_config = super(Attention1D, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class PseudoBlockSparseAttention1D(DynastesBaseLayer):

    def __init__(self,
                 num_heads,
                 block_size,
                 blocksparse_bijector: PseudoBlockSparseBijector1D,
                 multiquery_attention=False,
                 dropout_rate=0.,
                 mask_right=False,
                 **kwargs):
        super(PseudoBlockSparseAttention1D, self).__init__(**kwargs)
        self.num_heads = num_heads
        self.block_size = block_size
        self.multiquery_attention = multiquery_attention
        if self.multiquery_attention:
            self.num_heads_kv = 1
        else:
            self.num_heads_kv = num_heads
        self.blocksparse_bijector = blocksparse_bijector
        self.mask_right = mask_right
        self.dropout_rate = dropout_rate

    def call(self, inputs, training=None, mask=None):
        if len(inputs) == 3:
            q, k, v = inputs
        else:
            raise ValueError()

        if mask is not None:
            mask = mask[1]
            mask = tf.cast(mask, k.dtype)
            mask = tf.expand_dims(mask, axis=-1)
            mask_chain = self.blocksparse_bijector.get_bijector(mask)
            mask = mask_chain.forward(mask)
        causality_mat = None
        if self.mask_right:
            causality_mat = self.blocksparse_bijector.get_causality_matrix(k)
            causality_mat = tf.squeeze(causality_mat, axis=-1)

        q_chain = self.blocksparse_bijector.get_bijector(q)
        k_chain = self.blocksparse_bijector.get_bijector(k)
        v_chain = self.blocksparse_bijector.get_bijector(v)

        q = q_chain.forward(q)
        k = k_chain.forward(k)
        v = v_chain.forward(v)

        q = t2t_attention.split_heads(q, self.num_heads)
        k = t2t_attention.split_heads(k, self.num_heads_kv)
        v = t2t_attention.split_heads(v, self.num_heads_kv)
        if training:
            rate = self.dropout_rate
        else:
            rate = 0.
        r, weights = t2t_attention.masked_local_attention_1d(q=q, k=k, v=v, block_length=self.block_size,
                                                             mask_right=self.mask_right,
                                                             causality_tensor=causality_mat,
                                                             mask=tf.cast(mask, k.dtype),
                                                             dropout_rate=rate)

        r = t2t_attention.combine_heads(r)
        r_chain = self.blocksparse_bijector.get_bijector(r)
        return r_chain.inverse(r), None


def _get_attention_type_2D(local, masked, relative, self_attention, sparse):
    if self_attention:
        if sparse:
            raise NotImplementedError
        else:
            if local:
                # TODO: Add this
                """
                if masked:
                    attention_type = 'masked_local_attention_2d'
                else:
                    attention_type = 'unmasked_local_attention_2d'
                """
                raise NotImplementedError
            else:
                if relative:
                    if masked:
                        # attention_type = 'masked_self_attention_relative'
                        raise NotImplementedError
                    else:
                        attention_type = 'unmasked_self_attention_relative'
                else:
                    # attention_type = 'attention'
                    raise NotImplementedError
    else:
        raise NotImplementedError
    # TODO: Add these back?
    """
    if sparse:
        raise NotImplementedError
    if local:
        if masked:
            attention_type = 'masked_local_attention_2d'
        else:
            attention_type = 'unmasked_local_attention_2d'
    else:
        if relative:
            raise NotImplementedError
        else:
            attention_type = 'attention'
    """
    return attention_type


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class Attention2D(DynastesBaseLayer):

    def __init__(self,
                 num_heads,
                 multiquery_attention=False,
                 self_attention=True,
                 masked=False,
                 local=False,
                 relative=True,
                 sparse=False,
                 dropout_rate=0.0,
                 max_relative_position=None,
                 block_length=128,
                 filter_width=100,
                 mask_right=False,
                 add_relative_to_values=False,
                 heads_share_relative_embeddings=False,
                 **kwargs):
        super(Attention2D, self).__init__(**kwargs)

        attention_type = _get_attention_type_2D(local, masked, relative, self_attention, sparse)

        assert (attention_type in [
            'attention',
            'unmasked_self_attention_relative',
            'masked_self_attention_relative',
            'unmasked_local_attention_2d',
            'masked_local_attention_2d'])
        self.num_heads = num_heads
        self.multiquery_attention = multiquery_attention
        if self.multiquery_attention:
            self.num_heads_kv = 1
        else:
            self.num_heads_kv = num_heads
        self.attention_type = attention_type
        self.masked = masked
        self.local = local
        self.relative = relative
        self.sparse = sparse
        self.mask_right = mask_right
        self.dropout_rate = dropout_rate
        self.max_relative_position = max_relative_position
        self.block_length = block_length
        self.filter_width = filter_width
        self.add_relative_to_values = add_relative_to_values
        self.heads_share_relative_embeddings = heads_share_relative_embeddings

    def build(self, input_shape):
        self.depth_q = int(input_shape[0][-1]) // self.num_heads
        depth_k = int(input_shape[1][-1]) // self.num_heads_kv
        depth_v = int(input_shape[2][-1]) // self.num_heads_kv
        if 'relative' in self.attention_type:
            if self.attention_type == 'unmasked_self_attention_relative':
                if self.max_relative_position is None:
                    raise ValueError('max_relative_position cannot be None')
                k_embedding_shape = t2t_attention.get_relative_embeddings_left_right_shape(
                    self.max_relative_position,
                    depth_k, self.num_heads,
                    self.heads_share_relative_embeddings)
                self.height_key_embeddings = self.add_weight('height_key_embeddings',
                                                             initializer=tf.keras.initializers.RandomNormal(
                                                                 stddev=t2t_attention.get_embedding_initializer_stddev(
                                                                     depth_k)),
                                                             shape=k_embedding_shape)
                self.width_key_embeddings = self.add_weight('width_key_embeddings',
                                                            initializer=tf.keras.initializers.RandomNormal(
                                                                stddev=t2t_attention.get_embedding_initializer_stddev(
                                                                    depth_k)),
                                                            shape=k_embedding_shape)

                if self.add_relative_to_values:
                    raise ValueError("Adding relative embeddings to values is not implemented")
            elif self.attention_type == 'masked_self_attention_relative':
                # TODO: Maybe add this?
                """
                k_embedding_shape = t2t_attention.get_relative_embeddings_left_shape(
                    self.max_relative_position,
                    depth_k, self.num_heads_kv, False)
                self.key_embeddings = self.add_weight('key_embeddings',
                                                      initializer=tf.keras.initializers.RandomNormal(
                                                          stddev=t2t_attention.get_embedding_initializer_stddev(
                                                              depth_k)),
                                                      shape=k_embedding_shape)
                if self.add_relative_to_values:
                    v_embedding_shape = t2t_attention.get_relative_embeddings_left_shape(
                        self.max_relative_position,
                        depth_v, self.num_heads_kv,
                        False)
                    self.value_embeddings = self.add_weight('value_embeddings',
                                                            initializer=tf.keras.initializers.RandomNormal(
                                                                stddev=t2t_attention.get_embedding_initializer_stddev(
                                                                    depth_v)),
                                                            shape=v_embedding_shape)
                else:
                    self.value_embeddings = None
                """
                raise NotImplementedError
            else:
                raise ValueError()

    def call(self, inputs, training=None, mask=None):
        if len(inputs) == 3:
            q, k, v = inputs
        else:
            raise ValueError()

        # TODO: Add mask support
        # if mask is not None:
        #    q_mask = (1. - tf.cast(mask[0], tf.float32))[:, tf.newaxis, :, tf.newaxis]
        #    kv_mask = (1. - tf.cast(mask[1], tf.float32))[:, tf.newaxis, tf.newaxis, :]
        #    c_mask = tf.maximum(q_mask, kv_mask)
        #    if self.mask_right:
        #        look_ahead_mask = self._create_look_ahead_mask(mask[0].shape[1])
        #        c_mask = tf.maximum(c_mask, look_ahead_mask)
        #    bias = c_mask * large_compatible_negative(k.dtype)
        bias = None
        r = None
        weights = None

        q = t2t_attention.split_heads_2d(q, self.num_heads)
        k = t2t_attention.split_heads_2d(k, self.num_heads_kv)
        v = t2t_attention.split_heads_2d(v, self.num_heads_kv)

        if training:
            rate = self.dropout_rate
        else:
            rate = 0.

        if 'relative' in self.attention_type:
            height_key_embeddings = self.get_weight('height_key_embeddings', training=training)
            width_key_embeddings = self.get_weight('width_key_embeddings', training=training)
            if self.add_relative_to_values:
                raise ValueError("Adding relative embeddings to values is not implemented")
            if self.attention_type == 'unmasked_self_attention_relative':
                r, weights = t2t_attention.dot_product_unmasked_self_attention_relative_2d(q=q, k=k, v=v, bias=bias,
                                                                                           width_key_relative_embeddings=width_key_embeddings,
                                                                                           height_key_relative_embeddings=height_key_embeddings,
                                                                                           dropout_rate=rate,
                                                                                           max_relative_position=self.max_relative_position,
                                                                                           heads_share_relative_embedding=self.heads_share_relative_embeddings)
            else:
                raise ValueError("Not implemented")

            # TODO: does this case even make sense?

            # elif self.attention_type == 'masked_self_attention_relative':
            # r, weights = t2t_attention.dot_product_self_attention_relative_v2(q=q, k=k, v=v, bias=bias,
            #                                                                  key_left_embedding=key_embeddings,
            #                                                                  value_left_embedding=value_embeddings,
            #                                                                  dropout_rate=self.dropout_rate,
            #                                                                  max_relative_position=self.max_relative_position)
        else:
            raise ValueError("Not implemented")

        # TODO: Add these back
        """
        if self.attention_type == 'unmasked_local_attention_1d':
            r, weights = t2t_attention.local_attention_1d(q=q, k=k, v=v, block_length=self.block_length,
                                                          filter_width=self.filter_width)
        elif self.attention_type == 'masked_local_attention_1d':
            r, weights = t2t_attention.masked_local_attention_1d(q=q, k=k, v=v, block_length=self.block_length,
                                                                 dropout_rate=self.dropout_rate)
        elif self.attention_type == 'sparse_attention_truncated':
            raise ValueError("Not implemented")
        else:
            r, weights = t2t_attention.dot_product_attention(q=q, k=k, v=v, bias=bias,
                                                             dropout_rate=self.dropout_rate)
        """

        r = t2t_attention.combine_heads_2d(r)
        return r, weights

    def compute_output_shape(self, input_shape):
        depth_v = int(input_shape[2][-1]) // self.num_heads_kv
        return tuple(input_shape[:-1]) + (depth_v * self.num_heads,)

    def get_config(self):
        config = {
            'num_heads_q': self.num_heads_q,
            'masked': self.masked,
            'local': self.local,
            'relative': self.relative,
            'sparse': self.sparse,
            'dropout_rate': self.dropout_rate,
            'max_relative_position': self.max_relative_position,
            'block_length': self.block_length,
            'filter_width': self.filter_width,
            'mask_right': self.mask_right,
            'add_relative_to_values': self.add_relative_to_values,
            'heads_share_relative_embeddings': self.heads_share_relative_embeddings
        }
        base_config = super(Attention2D, self).get_config()
        return {**base_config, **config}



import tensorflow as tf

from dynastes.ops.t2t_common import shape_list


def _upscale2d(x, strides, method, antialias=True, gain=1):
    x_shape = shape_list(x)
    ret_h = x_shape[1] * strides[0]
    ret_w = x_shape[2] * strides[1]

    # Apply gain.
    if gain != 1:
        x *= gain

    return tf.image.resize(x, size=[ret_h, ret_w], method=method, antialias=antialias)


def _downscale2d(x, strides, method, antialias=True, gain=1):
    x_shape = shape_list(x)
    ret_h = x_shape[1] // strides[0]
    ret_w = x_shape[2] // strides[1]

    # Apply gain.
    if gain != 1:
        x *= gain

    return tf.image.resize(x, size=[ret_h, ret_w], method=method, antialias=antialias)


def upscale2d(x, strides=(2, 2), method='bilinear', antialias=True):
    with tf.name_scope('Upscale2D'):
        @tf.custom_gradient
        def func(x):
            y = _upscale2d(x, strides=strides, method=method, antialias=antialias)

            @tf.custom_gradient
            def grad(dy):
                with tf.name_scope('Upscale2DGrad'):
                    dx = _downscale2d(dy, strides=strides, method=method, antialias=antialias,
                                      gain=((strides[0] + strides[1]) / 2) ** 2)
                    return dx, lambda ddx: _upscale2d(ddx, strides=strides, method=method, antialias=antialias)

            return y, grad

        return func(x)


def downscale2d(x, strides=(2, 2), method='bilinear', antialias=True):
    with tf.name_scope('Downscale2D'):
        @tf.custom_gradient
        def func(x):
            y = _downscale2d(x, strides=strides, method=method, antialias=antialias)

            @tf.custom_gradient
            def grad(dy):
                with tf.name_scope('Downscale2DGrad'):
                    dx = _upscale2d(dy, strides=strides, method=method, antialias=antialias,
                                    gain=1 / ((strides[0] + strides[1]) / 2) ** 2)
                    return dx, lambda ddx: _downscale2d(ddx, strides=strides, method=method, antialias=antialias)

            return y, grad

        return func(x)



import numpy as np
import tensorflow as tf
import tensorflow.keras as tfk
import tensorflow.keras.layers as tfkl


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class SpectralNormalization(tfkl.Layer):
    def __init__(self,
                 power_iteration_rounds=1,
                 equality_constrained=True,
                 transposed=False,
                 trainable=None,
                 **kwargs):
        super(SpectralNormalization, self).__init__(**kwargs, trainable=False)
        self.power_iteration_rounds = power_iteration_rounds
        self.equality_constrained = equality_constrained
        self.transposed = transposed

    def build(self, input_shape):
        if self.transposed:
            input_shape = input_shape[:-2] + [input_shape[-1], input_shape[-2]]
        u_shape = (np.cumprod(np.array(input_shape[:-1]))[-1], 1)

        replica_context = tf.distribute.get_replica_context()
        if replica_context is None:  # cross repica strategy.
            # TODO(joelshor): Determine appropriate aggregation method.
            # https://github.com/tensorflow/gan/blob/master/tensorflow_gan/python/features/spectral_normalization.py
            raise ValueError("spectral norm isn't supported in cross-replica "
                             "distribution strategy.")
        elif not tf.distribute.has_strategy():  # default strategy.
            aggregation = None
        else:
            aggregation = tf.VariableAggregation.ONLY_FIRST_REPLICA

        self.u = self.add_weight(name='u',
                                 shape=u_shape,
                                 dtype=self.dtype,
                                 initializer=tfk.initializers.RandomNormal(),
                                 trainable=False,
                                 aggregation=aggregation)
        if self.transposed:
            input_shape = input_shape[:-2] + [input_shape[-1], input_shape[-2]]
        super(SpectralNormalization, self).build(input_shape)

    def _compute_spectral_norm(self, w, training=None):
        w = tf.reshape(w, (-1, w.get_shape()[-1]))
        u = self.u
        # Use power iteration method to approximate spectral norm.
        for i in range(self.power_iteration_rounds):
            # `v` approximates the first right singular vector of matrix `w`.
            v = tf.nn.l2_normalize(tf.matmul(a=w, b=u, transpose_a=True))
            u = tf.nn.l2_normalize(tf.matmul(w, v))

        # Update persisted approximation.
        if training:
            with tf.control_dependencies([self.u.assign(u, name='update_u')]):
                u = tf.identity(u)

        u = tf.stop_gradient(u)
        v = tf.stop_gradient(v)

        # Largest singular value of `w`.
        spectral_norm = tf.matmul(tf.matmul(a=u, b=w, transpose_a=True), v)
        spectral_norm.shape.assert_is_fully_defined()
        spectral_norm.shape.assert_is_compatible_with([1, 1])

        return spectral_norm[0][0]

    def call(self, w, training=None):
        if self.transposed:
            w = tf.linalg.matrix_transpose(w)
        normalization_factor = self._compute_spectral_norm(w, training=training)
        if not self.equality_constrained:
            normalization_factor = tf.maximum(1., normalization_factor)
        w_normalized = w / normalization_factor
        w_normalized = tf.reshape(w_normalized, w.get_shape())
        if self.transposed:
            w_normalized = tf.linalg.matrix_transpose(w_normalized)
        return w_normalized

    def get_config(self):
        config = {
            'power_iteration_rounds': self.power_iteration_rounds,
            'equality_constrained': self.equality_constrained,
            'transposed': self.transposed
        }
        base_config = super(SpectralNormalization, self).get_config()
        return {**base_config, **config}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from functools import partial

import tensorflow as tf

from dynastes import activations
from dynastes.blocks import layer_factory
from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.ops import t2t_common
from dynastes.util import cache_context
from dynastes.util.layer_util import call_masked as cm
from dynastes.util.layer_util import compute_mask_if_possible as compm


# A module that only depends on `keras.layers` import these from here.

@tf.keras.utils.register_keras_serializable(package='Dynastes')
class _AttentionBlock1D(DynastesBaseLayer):

    def __init__(self,
                 attention_dim,
                 output_dim,
                 kernel_size=1,
                 attention_type='Attention1D',
                 q_type='Conv1D',
                 k_type=None,
                 v_type=None,
                 num_heads=1,
                 multiquery_attention=False,
                 depth_multiplier=1,
                 strides=1,
                 dilation_rate=1,
                 grouped=False,
                 group_size=1,
                 padding='same',
                 activation=None,
                 use_bias=False,
                 relative=False,
                 local=False,
                 sparse=False,
                 masked=False,
                 dropout_rate=0.,
                 max_relative_position=2,
                 lsh_bucket_length=4,
                 block_length=4,
                 filter_width=2,
                 mask_right=False,
                 add_relative_to_values=False,
                 return_attn_weights=False,
                 cache_kv=False,
                 **kwargs):
        kwargs['supports_caching'] = True
        super(_AttentionBlock1D, self).__init__(**kwargs)
        self.q_type = q_type
        self.activation = activations.get(activation)
        self.use_bias = use_bias

        if k_type is None:
            self.k_type = q_type
        else:
            self.k_type = k_type

        if v_type is None:
            self.v_type = q_type
        else:
            self.v_type = v_type

        self.attention_dim = attention_dim
        self.output_dim = output_dim
        self.kernel_size = kernel_size
        self.num_heads = num_heads
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.depth_multiplier = depth_multiplier
        self.padding = padding
        self.grouped = grouped
        self.group_size = group_size
        self.multiquery_attention = multiquery_attention
        self.relative = relative
        self.local = local
        self.masked = masked
        self.sparse = sparse
        self.dropout_rate = dropout_rate
        self.max_relative_position = max_relative_position
        self.lsh_bucket_length = lsh_bucket_length
        self.block_length = block_length
        self.filter_width = filter_width
        self.mask_right = mask_right
        self.add_relative_to_values = add_relative_to_values
        self.return_attn_weights = return_attn_weights
        self.supports_masking = True
        self.cache_kv = cache_kv
        conv_partial = partial(layer_factory.get_1d_layer, kernel_size=kernel_size,
                               grouped=grouped,
                               group_size=group_size,
                               depth_multiplier=depth_multiplier,
                               padding=padding,
                               activation=activation,
                               use_bias=use_bias,
                               kernel_initializer=self.get_initializer('kernel'),
                               bias_initializer=self.get_initializer('bias'),
                               kernel_regularizer=self.get_regularizer('kernel'),
                               bias_regularizer=self.get_regularizer('bias'),
                               activity_regularizer=None,
                               kernel_constraint=self.get_constraint('kernel'),
                               bias_constraint=self.get_constraint('bias'))
        q_filters = attention_dim
        k_filters = attention_dim
        v_filters = output_dim

        if multiquery_attention:
            k_filters //= num_heads
            v_filters //= num_heads
        self.k_filters = k_filters
        self.v_filters = v_filters

        q_strides = strides
        kv_strides = strides
        kv_dilation_rate = dilation_rate
        attn_strides = 1
        attn_dilation_rate = 1
        if attention_type in ['LocalizedAttentionLayer1D'] and strides != 1:
            kv_strides = 1
            attn_strides = strides
            attn_dilation_rate = dilation_rate

        self.q_layer = conv_partial(type=self.q_type,
                                    filters=q_filters,
                                    strides=q_strides,
                                    dilation_rate=dilation_rate, name='Conv-Q')
        self.k_layer = conv_partial(type=self.k_type,
                                    filters=k_filters,
                                    strides=kv_strides,
                                    dilation_rate=kv_dilation_rate, name='Conv-K')
        self.v_layer = conv_partial(type=self.v_type,
                                    filters=v_filters,
                                    strides=kv_strides,
                                    dilation_rate=kv_dilation_rate, name='Conv-V')

        attention_padding = padding
        self.attention_layer = layer_factory.get_1D_attention_layer(
            type=attention_type,
            strides=attn_strides,
            dilation_rate=attn_dilation_rate,
            num_heads=num_heads,
            padding=attention_padding,
            multiquery_attention=multiquery_attention,
            preshaped_q=True,
            relative=relative,
            local=local,
            masked=masked,
            sparse=sparse,
            dropout_rate=dropout_rate,
            max_relative_position=max_relative_position,
            lsh_bucket_length=lsh_bucket_length,
            block_length=block_length,
            mask_right=mask_right,
            filter_width=filter_width,
            add_relative_to_values=add_relative_to_values,
        )

    def compute_mask(self, inputs, mask=None):
        if mask is not None:
            q_mask = compm(self.q_layer, inputs, mask=mask)
            return q_mask
        return mask

    def get_config(self):
        config = {
            'q_type': self.q_type,
            'k_type': self.k_type,
            'v_type': self.v_type,
            'activation': activations.serialize(self.activation),
            'use_bias': self.use_bias,
            'multiquery_attention': self.multiquery_attention,
            'attention_dim': self.attention_dim,
            'output_dim': self.output_dim,
            'depth_multiplier': self.depth_multiplier,
            'kernel_size': self.kernel_size,
            'num_heads': self.num_heads,
            'strides': self.strides,
            'dilation_rate': self.dilation_rate,
            'padding': self.padding,
            'grouped': self.grouped,
            'group_size': self.group_size,
            'relative': self.relative,
            'local': self.local,
            'masked': self.masked,
            'sparse': self.sparse,
            'dropout_rate': self.dropout_rate,
            'max_relative_position': self.max_relative_position,
            'lsh_bucket_length': self.lsh_bucket_length,
            'block_length': self.block_length,
            'filter_width': self.filter_width,
            'mask_right': self.mask_right,
            'add_relative_to_values': self.add_relative_to_values,
            'cache_kv': self.cache_kv,
        }
        base_config = super(_AttentionBlock1D, self).get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shape):
        qs = self.q_layer.compute_output_shape(input_shape)
        ks = self.k_layer.compute_output_shape(input_shape)
        vs = self.v_layer.compute_output_shape(input_shape)
        output_shape = self.attention_layer.compute_output_shape([qs, ks, vs])
        if self.return_attn_weights:
            return output_shape
        return output_shape[0]


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class AttentionBlock1D(_AttentionBlock1D):

    def request_cache(self, batch_size=1, max_length=1):
        return {
        }

    def compute_mask(self, inputs, mask=None):
        if mask is not None:
            q_mask = compm(self.q_layer, inputs[0], mask=mask[0])
            return q_mask
        return mask

    def call(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None):
        qx, sx = inputs
        if mask is not None:
            qmask, smask = mask
        else:
            qmask, smask = (None, None)
        q, q_mask = cm(self.q_layer, qx, training=training, mask=qmask)

        # In cross-attention we may compute KV once and perform attention on it in subsequent steps

        def get_kv():
            def do_get_kv():
                k, k_mask = cm(self.k_layer, sx, training=training, mask=smask)
                v, v_mask = cm(self.v_layer, sx, training=training, mask=smask)
                return k, k_mask, v, v_mask

            if self.cache_kv and cache_context.cache_context is not None:
                with cache_context.SubContext(self.name):
                    if 'cached_kv' in cache_context.cache:
                        k, k_mask, v, v_mask = cache_context.cache['cached_kv']
                    else:
                        k, k_mask, v, v_mask = do_get_kv()
                        cache_context.cache['cached_kv'] = k, k_mask, v, v_mask
            else:
                k, k_mask, v, v_mask = do_get_kv()
            return k, k_mask, v, v_mask

        if cache is not None:
            if 'k' not in cache:
                k, k_mask, v, v_mask = get_kv()

                # Update cache
                cache["k"] = k
                cache["v"] = v
                if mask is not None:
                    cache["k_mask"] = k_mask
                    cache["v_mask"] = v_mask
            else:
                k = cache["k"]
                v = cache["v"]
                if mask is not None:
                    k_mask = cache["k_mask"]
                    v_mask = cache["v_mask"]
        else:
            k, k_mask, v, v_mask = get_kv()
        if mask is not None:
            mask = [q_mask, tf.logical_and(k_mask, v_mask)]
        x, weights = self.attention_layer([q, k, v], mask=mask, training=training)

        if self.return_attn_weights:
            return x, weights
        return x


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class SelfAttentionBlock1D(AttentionBlock1D):

    def request_cache(self, batch_size=1, max_length=1):
        return {
            'k': tf.zeros((batch_size, max_length, self.k_filters)),
            'v': tf.zeros((batch_size, max_length, self.v_filters)),
            'k_mask': tf.cast(tf.zeros((batch_size, max_length)), tf.bool),
            'v_mask': tf.cast(tf.zeros((batch_size, max_length)), tf.bool)
        }

    def compute_mask(self, inputs, mask=None):
        if mask is not None:
            q_mask = compm(self.q_layer, inputs, mask=mask)
            return q_mask
        return mask

    def call(self, inputs, training=None, mask=None, cache=None, decode_loop_step=None, pad_q_to_kv=False):
        x = inputs
        q, q_mask = cm(self.q_layer, x, training=training, mask=mask)
        k, k_mask = cm(self.k_layer, x, training=training, mask=mask)
        v, v_mask = cm(self.v_layer, x, training=training, mask=mask)
        if cache is not None:
            # Combine cached keys and values with new keys and values.
            if cache["k"] is not None:
                # Update cache
                if decode_loop_step is not None:

                    cache_k_shape = cache["k"].shape.as_list()
                    indices = tf.reshape(
                        tf.one_hot(decode_loop_step, cache_k_shape[1], dtype=k.dtype),
                        [1, cache_k_shape[1], 1])
                    k = cache["k"] + k * indices
                    if mask is not None:
                        indices = tf.reshape(
                            tf.one_hot(decode_loop_step, cache_k_shape[1], dtype=tf.float16),
                            [1, cache_k_shape[1]])
                        k_mask = tf.logical_or(cache["k_mask"], (tf.cast(k_mask, tf.float16) * indices) > 0.)

                    cache_v_shape = cache["v"].shape.as_list()
                    indices = tf.reshape(
                        tf.one_hot(decode_loop_step, cache_v_shape[1], dtype=v.dtype),
                        [1, cache_v_shape[1], 1])
                    v = cache["v"] + v * indices
                    if mask is not None:
                        indices = tf.reshape(
                            tf.one_hot(decode_loop_step, cache_v_shape[1], dtype=tf.float16),
                            [1, cache_v_shape[1]])
                        v_mask = tf.logical_or(cache["v_mask"], (tf.cast(v_mask, tf.float16) * indices) > 0.)
                else:
                    k = tf.concat([tf.cast(cache["k"], k.dtype), k], axis=1)
                    v = tf.concat([tf.cast(cache["v"], v.dtype), v], axis=1)
                    if mask is not None:
                        k_mask = tf.concat([tf.cast(cache["k_mask"], k_mask.dtype), k_mask], axis=1)
                        v_mask = tf.concat([tf.cast(cache["v_mask"], v_mask.dtype), v_mask], axis=1)

            # Update cache
            cache["k"] = k
            cache["v"] = v
            if mask is not None:
                cache["k_mask"] = k_mask
                cache["v_mask"] = v_mask

        q_shape = t2t_common.shape_list(q)
        kv_shape = t2t_common.shape_list(k)

        if pad_q_to_kv:
            if q_shape[1] != kv_shape[1]:
                if decode_loop_step is not None:
                    q_prepad = decode_loop_step
                    q_postpad = (kv_shape[1] - q_shape[1]) - decode_loop_step

                else:
                    q_prepad = (kv_shape[1] - q_shape[1])
                    q_postpad = 0
                q = tf.pad(q, paddings=[[0, 0], [q_prepad, q_postpad], [0, 0]])
                if mask is not None:
                    q_mask = tf.pad(q_mask, paddings=[[0, 0], [q_prepad, q_postpad]])
            else:
                # This is just stupid autograph nonsense, ignore it
                if decode_loop_step is not None:
                    q_prepad = decode_loop_step
                else:
                    q_prepad = (kv_shape[1] - q_shape[1])
        else:
            # This is just stupid autograph nonsense, ignore it
            if decode_loop_step is not None:
                q_prepad = decode_loop_step
            else:
                q_prepad = (kv_shape[1] - q_shape[1])

        if mask is not None:
            mask = [q_mask, tf.logical_and(k_mask, v_mask)]
        x, weights = self.attention_layer([q, k, v], mask=mask, training=training)
        x_shape = t2t_common.shape_list(x)
        if pad_q_to_kv:
            if q_shape[1] != kv_shape[1]:
                if decode_loop_step is not None:
                    x = tf.slice(x, [0, q_prepad, 0], [x_shape[0], 1, x_shape[2]])
                else:
                    x = tf.slice(x, [0, q_prepad, 0], [x_shape[0], q_shape[1], x_shape[2]])
        if self.return_attn_weights:
            return x, weights
        return x



import numpy as np
import tensorflow as tf
from tensorflow.python.framework import test_util

from dynastes.layers.attention_layers import LocalizedAttentionLayer1D, LocalizedAttentionLayer2D


def _test_grads(testCase: tf.test.TestCase, func, input):
    _, grads = tf.test.compute_gradient(func, input)
    for grad in grads:
        testCase.assertNotAllClose(grad, np.zeros_like(grad))


to_tensor = tf.convert_to_tensor
normal = np.random.normal


class LocalizedAttentionLayer1DTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        t_steps = 12
        bs = 2
        dim = 16
        v_dim = dim * 2
        s = 2

        layer = LocalizedAttentionLayer1D(kernel_size=3,
                                          strides=s,
                                          num_heads=4,
                                          dilation_rate=2)

        q = to_tensor(normal(size=(bs, t_steps // s, dim))
                      .astype(np.float32))
        k = to_tensor(normal(size=(bs, t_steps, dim))
                      .astype(np.float32))
        v = to_tensor(normal(size=(bs, t_steps, v_dim))
                      .astype(np.float32))

        @tf.function
        def test_func(q, k, v):
            r, _ = layer(q, k, v)
            return r

        r = test_func(q, k=k, v=v)

        ex_res_shape = np.zeros((bs, t_steps // s, v_dim))

        self.assertShapeEqual(ex_res_shape, r)

        _test_grads(self, test_func, [q, k, v])


class LocalizedAttentionLayer2DTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        in_shape = [4, 4]
        bs = 1
        heads = 2
        dim = 2
        v_dim = dim * 2
        s = 2

        layer = LocalizedAttentionLayer2D(kernel_size=(3, 3),
                                          strides=(s, s),
                                          num_heads=heads,
                                          dilation_rate=(1, 1))

        q = to_tensor(normal(size=(bs, in_shape[0] // s, in_shape[1] // s, dim * heads))
                      .astype(np.float32))
        k = to_tensor(normal(size=(bs, in_shape[0], in_shape[1], dim * heads))
                      .astype(np.float32))
        v = to_tensor(normal(size=(bs, in_shape[0], in_shape[1], v_dim * heads))
                      .astype(np.float32))

        @tf.function
        def test_func(_q, _k, _v):
            return layer(_q, _k, _v)

        r = test_func(q, _k=k, _v=v)

        ex_res_shape = np.zeros((bs, in_shape[0] // s, in_shape[1] // s, v_dim * heads))

        self.assertShapeEqual(ex_res_shape, r)

        _test_grads(self, test_func, [q, k, v])

    @test_util.use_deterministic_cudnn
    def test_multiquery(self):
        in_shape = [4, 4]
        bs = 1
        heads = 2
        dim = 2
        v_dim = dim * 2
        s = 2

        layer = LocalizedAttentionLayer2D(kernel_size=(3, 3),
                                          strides=(s, s),
                                          num_heads=heads,
                                          dilation_rate=(1, 1), multiquery_attention=True)

        q = to_tensor(normal(size=(bs, in_shape[0] // s, in_shape[1] // s, dim * heads))
                      .astype(np.float32))
        k = to_tensor(normal(size=(bs, in_shape[0], in_shape[1], dim))
                      .astype(np.float32))
        v = to_tensor(normal(size=(bs, in_shape[0], in_shape[1], v_dim))
                      .astype(np.float32))

        @tf.function
        def test_func(_q, _k, _v):
            return layer(_q, k=_k, v=_v)

        r = test_func(q, _k=k, _v=v)

        ex_res_shape = np.zeros((bs, in_shape[0] // s, in_shape[1] // s, v_dim * heads))

        self.assertShapeEqual(ex_res_shape, r)

        _test_grads(self, test_func, [q, k, v])


if __name__ == '__main__':
    tf.config.optimizer.set_jit(True)
    tf.test.main()



import abc

import numpy as np
import tensorflow as tf

from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.ops.t2t_common import shape_list


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class ModulationLayer(DynastesBaseLayer, abc.ABC):

    def compute_output_shape(self, input_shape):
        assert isinstance(input_shape, list)
        return input_shape[0]


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class FeaturewiseLinearModulation(ModulationLayer):
    """
        Call accepts list of:
            * [input, modulation]
            * [input, modulation bias, modulation scale]
            * [input, modulation bias + modulation scale]

        Modulation can have smaller dimensions than input,
        as long as '''rank(modulation) in [rank(input), 2]'''
            input: [b, H, W, CH]
                modulation:
                    [b, H, W, N] - Valid
                    [b, N] - Valid
                    [b, other H, other W, N] - Valid
                    [b, any, N] - NOT VALID

        More information
        https://distill.pub/2018/feature-wise-transformations/
    """

    def __init__(self, method=tf.image.ResizeMethod.BILINEAR, antialias=True, mode=None, **kwargs):

        """
        @param method: Method used for scaling conditioning in case rank(input) == rank(modulation)
        @type method: tensorflow.image.ResizeMethod
        @param antialias: Use antialiasing if downscaling conditioning
        @type antialias:
        @param mode: one of 'provided_mean_var', 'mapped' or 'provided_meanvar_fused',
                     leave blank to infer on build
        @type mode: str
        """
        super(FeaturewiseLinearModulation, self).__init__(**kwargs)
        self.method = method
        self.antialias = antialias
        self.mode = mode

    def build(self, input_shape):
        assert isinstance(input_shape, list)
        if len(input_shape) == 3:
            assert (input_shape[1][-1] + input_shape[2][-1]) // 2 == input_shape[0][-1]
            if self.mode is not None:
                assert self.mode == 'provided_mean_var'
            else:
                self.mode = 'provided_mean_var'
        elif len(input_shape) == 2:
            if (input_shape[1][-1] // 2) != input_shape[0][-1]:
                if self.mode is not None:
                    assert self.mode == 'mapped'
                else:
                    self.mode = 'mapped'
                self.add_weight('map_kernel', shape=[input_shape[1][-1], input_shape[0][-1] * 2])
            else:
                if self.mode is not None:
                    assert self.mode == 'provided_meanvar_fused'
                else:
                    self.mode = 'provided_meanvar_fused'
        else:
            raise ValueError('Incorrect input shapes')
        super(FeaturewiseLinearModulation, self).build(input_shape)

    def call(self, x, training=None, mask=None):
        assert isinstance(x, list)
        if self.mode == 'provided_mean_var':
            x, mean, var = x
            mean_var = [mean, var]
            mean_var = tf.concat([mean_var], axis=-1)
        elif self.mode == 'mapped':
            x, mean_var = x
            mean_var = tf.matmul(mean_var, self.get_weight('map_kernel', training=training))
        elif self.mode != 'provided_meanvar_fused':
            raise ValueError('Something is wrong')
        else:
            x, mean_var = x
        mean_var_shape = shape_list(mean_var)
        x_shape = shape_list(x)
        if len(mean_var_shape) != len(x_shape):
            mean_var = tf.reshape(mean_var, [-1, ] + ([1] * (len(x.shape) - 2)) + [2, x_shape[-1]])
        else:
            mean_var_shape_npa = np.array(mean_var_shape[1:-1])
            x_shape_npa = np.array(x_shape[1:-1])
            compatible = np.all(np.logical_or(mean_var_shape_npa == 1, mean_var_shape_npa == x_shape_npa))
            if not compatible:
                size = np.where(mean_var_shape_npa == 1, mean_var_shape_npa, x_shape_npa).tolist()
                if len(mean_var_shape) == 4:
                    mean_var = tf.image.resize(mean_var, size, method=self.method)
                elif len(mean_var_shape) == 3:
                    mean_var = tf.squeeze(tf.image.resize(tf.expand_dims(mean_var, 1), [1] + size, method=self.method,
                                                          antialias=self.antialias), axis=1)
                else:
                    raise ValueError('Only works for 1D or 2D tensors')

            shape = [mean_var_shape[0]] + np.where(mean_var_shape_npa == 1, mean_var_shape_npa,
                                                   x_shape_npa).tolist() + [2, x_shape[-1]]
            mean_var = tf.reshape(mean_var, shape)
        mean, var = tf.unstack(mean_var, axis=-2)

        return (x * var) + mean

    def get_config(self):
        config = {
            'method': self.method,
            'antialias': self.antialias,
            'mode': self.mode,
        }
        base_config = super(FeaturewiseLinearModulation, self).get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shape):
        assert isinstance(input_shape, list)
        return input_shape[0]



import tensorflow as tf
from tensorflow.keras.regularizers import Regularizer
from tensorflow.python.keras import regularizers as tf_regularizers
from tensorflow.python.keras.regularizers import deserialize as _deserialize
from tensorflow.python.keras.regularizers import serialize as _serialize

from . import orthogonal
from .orthogonal import Orthogonal


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class ModifyingRegularizer(Regularizer):

    def __init__(self, first, second,
                 fn,
                 **kwargs):
        super(ModifyingRegularizer).__init__(**kwargs)
        self.first = get(first)
        self.second = get(second)
        self.fn = fn

    def __call__(self, x):
        return self.fn(self.first(x), self.second(x))

    def get_config(self):
        return {'first': serialize(self.first),
                'second': serialize(self.second)}


def _add(self, other):
    return ModifyingRegularizer(self, other, lambda x, y: x + y)


def _sub(self, other):
    return ModifyingRegularizer(self, other, lambda x, y: x - y)


def _mul(self, other):
    return ModifyingRegularizer(self, other, lambda x, y: x * y)


setattr(Regularizer, '__add__', _add)
setattr(Regularizer, '__sub__', _sub)
setattr(Regularizer, '__mul__', _mul)


def serialize(regularizer):
    return _serialize(regularizer)


def deserialize(config, custom_objects={}):
    return _deserialize(config, custom_objects)


def get(regularizer):
    if type(regularizer) == str:
        if regularizer == 'orthogonal':
            return Orthogonal()
    return tf_regularizers.get(regularizer)



import tensorflow as tf
from tensorflow_addons.image import sparse_image_warp

from dynastes.ops.t2t_common import shape_list


def sparse_warp(mel_spectrograms, time_warping_para: float = 80.):
    """Spec augmentation Calculation Function.
    'SpecAugment' have 3 steps for audio data augmentation.
    first step is time warping using Tensorflow's image_sparse_warp function.
    Second step is frequency masking, last step is time masking.
    Args:
      mel_spectrograms: Tensor of log magnitudes and possibly instantaneous frequencies,
            shape [..., time, freq, ch*(1/2)], mel scaling of frequencies.
      time_warping_para(float): Augmentation parameter, "time warp parameter W".
        If none, default = 80 for LibriSpeech.
    Returns:
      mel_spectrograms: Tensor of log magnitudes and possibly instantaneous frequencies,
            shape [..., time, freq, ch*(1/2)], mel scaling of frequencies.
    """

    fbank_size = shape_list(mel_spectrograms)
    _, n, n_mels, _ = fbank_size
    # Step 1 : Time warping
    # Image warping control point setting.
    # Source
    pt = tf.random.uniform([], 0, n - (time_warping_para * 2),
                           tf.float32) + time_warping_para  # radnom point along the time axis
    src_ctr_pt_freq = tf.cast(tf.range(n_mels // 2), tf.float32)  # control points on freq-axis
    src_ctr_pt_time = tf.ones_like(src_ctr_pt_freq) * pt  # control points on time-axis
    src_ctr_pts = tf.stack((src_ctr_pt_time, src_ctr_pt_freq), -1)
    src_ctr_pts = tf.cast(src_ctr_pts, dtype=tf.float32)

    # Destination
    w = tf.random.uniform([], -time_warping_para, time_warping_para, tf.float32)  # distance
    dest_ctr_pt_freq = src_ctr_pt_freq
    dest_ctr_pt_time = src_ctr_pt_time + w
    dest_ctr_pts = tf.stack((dest_ctr_pt_time, dest_ctr_pt_freq), -1)
    dest_ctr_pts = tf.cast(dest_ctr_pts, dtype=tf.float32)

    # warp
    source_control_point_locations = tf.expand_dims(src_ctr_pts, 0)  # (1, v//2, 2)
    dest_control_point_locations = tf.expand_dims(dest_ctr_pts, 0)  # (1, v//2, 2)
    warped_image, _ = sparse_image_warp(mel_spectrograms,
                                        source_control_point_locations,
                                        dest_control_point_locations)
    return warped_image


def frequency_masking(mel_spectrograms, frequency_masking_para: int = 100, frequency_mask_num: int = 1, roll_mask=None):
    """Spec augmentation Calculation Function.
    'SpecAugment' have 3 steps for audio data augmentation.
    first step is time warping using Tensorflow's image_sparse_warp function.
    Second step is frequency masking, last step is time masking.
    Args:
      mel_spectrograms: Tensor of log magnitudes and possibly instantaneous frequencies,
            shape [..., time, freq, ch*(1/2)], mel scaling of frequencies.
      frequency_masking_para(int): Augmentation parameter, "frequency mask parameter F"
        If none, default = 100 for LibriSpeech.
      frequency_mask_num(int): number of frequency masking lines, "m_F".
        If none, default = 1 for LibriSpeech.
    Returns:
      mel_spectrograms: Tensor of log magnitudes and possibly instantaneous frequencies,
            shape [..., time, freq, ch*(1/2)], mel scaling of frequencies.
    """
    # Step 2 : Frequency masking
    fbank_size = shape_list(mel_spectrograms)
    _, n, n_mels, _ = fbank_size
    frequency_masking_para = min(frequency_masking_para, n_mels // 2)

    for i in range(frequency_mask_num):
        f = tf.random.uniform([], minval=0, maxval=frequency_masking_para, dtype=tf.int32)
        f0 = tf.random.uniform([], minval=0, maxval=n_mels - f, dtype=tf.int32)

        # warped_mel_spectrogram[f0:f0 + f, :] = 0
        mask = tf.concat((tf.ones(shape=(1, n, n_mels - f0 - f, 1)),
                          tf.zeros(shape=(1, n, f, 1)),
                          tf.ones(shape=(1, n, f0, 1)),
                          ), 2)
        if roll_mask is not None:
            roll_mel_spectrograms = tf.roll(mel_spectrograms, roll_mask, axis=0)
            mel_spectrograms = (mel_spectrograms * mask) + (roll_mel_spectrograms * (1-mask))
        else:
            mel_spectrograms = mel_spectrograms * mask
    return tf.cast(mel_spectrograms, dtype=tf.float32)


def time_masking(mel_spectrograms, time_masking_para: int = 27, time_mask_num: int = 1, roll_mask=None):
    """Spec augmentation Calculation Function.
    'SpecAugment' have 3 steps for audio data augmentation.
    first step is time warping using Tensorflow's image_sparse_warp function.
    Second step is frequency masking, last step is time masking.
    Args:
      mel_spectrograms(tf.Tensor): Tensor of log magnitudes and possibly instantaneous frequencies / phases,
            shape [..., time, freq, ch*(1/2)], mel scaling of frequencies.
      time_masking_para(int): Augmentation parameter, "time mask parameter T"
        If none, default = 27 for LibriSpeech.
      time_mask_num(int): number of time masking lines, "m_T".
        If none, default = 1 for LibriSpeech.
    Returns:
      mel_spectrogram: Tensor of log magnitudes and possibly instantaneous frequencies,
            shape [..., time, freq, ch*(1/2)], mel scaling of frequencies.
    """
    fbank_size = shape_list(mel_spectrograms)
    _, n, n_mels, _ = fbank_size
    # Step 3 : Time masking
    for i in range(time_mask_num):
        t = tf.random.uniform([], minval=0, maxval=time_masking_para, dtype=tf.int32)
        t0 = tf.random.uniform([], minval=0, maxval=n - t, dtype=tf.int32)

        # mel_spectrograms[:, t0:t0 + t] = 0
        mask = tf.concat((tf.ones(shape=(1, n - t0 - t, n_mels, 1)),
                          tf.zeros(shape=(1, t, n_mels, 1)),
                          tf.ones(shape=(1, t0, n_mels, 1)),
                          ), 1)
        if roll_mask is not None:
            roll_mel_spectrograms = tf.roll(mel_spectrograms, roll_mask, axis=0)
            mel_spectrograms = (mel_spectrograms * mask) + (roll_mel_spectrograms * (1-mask))
        else:
            mel_spectrograms = mel_spectrograms * mask

    return tf.cast(mel_spectrograms, dtype=tf.float32)


def spec_augment(mel_spectrograms: tf.Tensor,
                 time_warping_para: float = 80.,
                 time_masking_para: int = 27,
                 time_mask_num: int = 1,
                 frequency_masking_para: int = 100,
                 frequency_mask_num: int = 1,
                 normalize=True,
                 roll_mask=None):
    """
    Args:
      mel_spectrograms(tf.Tensor): Tensor of log magnitudes and possibly instantaneous frequencies / phases,
            shape [..., time, freq, ch*(1/2)], mel scaling of frequencies.
      time_warping_para(float): Augmentation parameter, "time warp parameter W".
        If none, default = 80 for LibriSpeech.
      time_masking_para(int): Augmentation parameter, "time mask parameter T"
        If none, default = 27 for LibriSpeech.
      time_mask_num(int): number of time masking lines, "m_T".
        If none, default = 1 for LibriSpeech.
      frequency_masking_para(int): Augmentation parameter, "frequency mask parameter F"
        If none, default = 100 for LibriSpeech.
      frequency_mask_num(int): number of frequency masking lines, "m_F".
        If none, default = 1 for LibriSpeech.
      roll_mask(int): if not none, replace masked indices with batch rolled by [roll_mask] steps
    Returns:
      mel_spectrograms: Tensor of log magnitudes and possibly instantaneous frequencies,
            shape [..., time, freq, ch*(1/2)], mel scaling of frequencies.
    """
    if normalize:
        mean, std = tf.nn.moments(mel_spectrograms, axes=[0, 1], keepdims=True)
        mel_spectrograms = (mel_spectrograms - mean) / std

    warped_mel_spectrogram = sparse_warp(mel_spectrograms,
                                         time_warping_para=time_warping_para)

    warped_frequency_spectrogram = frequency_masking(warped_mel_spectrogram,
                                                     frequency_masking_para=frequency_masking_para,
                                                     frequency_mask_num=frequency_mask_num,
                                                     roll_mask=roll_mask)

    warped_frequency_time_spectrogram = time_masking(warped_frequency_spectrogram,
                                                     time_masking_para=time_masking_para,
                                                     time_mask_num=time_mask_num,
                                                     roll_mask=roll_mask)

    if normalize:
        warped_frequency_time_spectrogram = (warped_frequency_time_spectrogram * std) + mean
    return warped_frequency_time_spectrogram



import math

import tensorflow as tf

from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.ops import spectral_ops


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class Wave2STFTLayer(DynastesBaseLayer):

    def __init__(self,
                 n_bins,
                 hop_length,
                 padding,
                 keep_dc=False,
                 hq=True,
                 **kwargs):
        super(Wave2STFTLayer, self).__init__(**kwargs)
        self.n_bins = n_bins
        self.hop_length = hop_length
        self.padding = padding
        self.keep_dc = keep_dc
        self.hq = hq
        self.inverted = False

    def call(self, inputs, invert=False, **kwargs):

        if invert or self.inverted:
            stfts = inputs
            if self.hq:
                stfts = tf.cast(stfts, tf.complex128)
            waves = spectral_ops.stfts_to_waves(stfts,
                                                n_fft=self.n_bins * 2,
                                                hop_length=self.hop_length,
                                                pad_l=self.padding,
                                                pad_r=self.padding,
                                                discard_dc=not self.keep_dc)
            return waves

        stfts = spectral_ops.waves_to_stfts(inputs,
                                            n_fft=self.n_bins * 2,
                                            hop_length=self.hop_length,
                                            pad_l=self.padding,
                                            pad_r=self.padding,
                                            discard_dc=not self.keep_dc,
                                            hq=self.hq)
        return stfts#tf.cast(stfts, inputs.dtype)

    def compute_output_shape(self, input_shape, invert=False):
        if invert:
            self.inverted = True
            output_shape = super(Wave2STFTLayer, self).compute_output_shape(input_shape)
            self.inverted = False
            return output_shape
        out_bins = self.n_bins
        if self.keep_dc:
            out_bins += 1
        print(input_shape)
        out_len = max(0, math.floor(((input_shape[1] + (self.padding * 2) - (self.n_bins * 2)) / self.hop_length)) + 1)
        output_shape = [input_shape[0], out_len, out_bins, input_shape[-1]]
        print(output_shape)
        return tf.TensorShape(dims=output_shape)

    def compute_output_signature(self, input_signature: tf.TensorSpec, invert=False):
        if invert:
            self.inverted = True
            output_signature = super(Wave2STFTLayer, self).compute_output_signature(input_signature)
            self.inverted = False
            return output_signature
        else:
            out_bins = self.n_bins
            if self.keep_dc:
                out_bins += 1
            output_dtype = tf.complex64
            if self.hq:
                output_dtype = tf.complex128
            input_shape = input_signature.shape.as_list()
            out_len = max(0, math.floor(((input_shape[1] + (self.padding * 2) - (self.n_bins * 2)) / self.hop_length)) + 1)
            output_shape = tf.TensorShape([input_shape[0], out_len, out_bins, input_shape[2]])
            return tf.TensorSpec(shape=output_shape, dtype=output_dtype)

    def get_config(self):
        config = {'n_bins': self.n_bins,
                  'hop_length': self.hop_length,
                  'padding': self.padding,
                  'keep_dc': self.keep_dc,
                  'hq': self.hq}
        base_config = super(Wave2STFTLayer, self).get_config()
        return {**base_config, **config}


def _get_lower_and_upper_edges_l2mel(min_hz, max_hz, n_mels):
    mxhzl2 = math.log2(max_hz)
    mnhzl2 = math.log2(min_hz)

    l2pmel = (mxhzl2 - mnhzl2) / n_mels

    low_edge_l2 = mnhzl2 - (l2pmel / 2)
    high_edge_l2 = mxhzl2 + (l2pmel / 2)

    low_edge_hz = 2 ** low_edge_l2
    high_edge_hz = 2 ** high_edge_l2
    return low_edge_hz, high_edge_hz


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class _l2melMatrixInitializer(tf.keras.initializers.Initializer):

    def __init__(self,
                 min_hz=40.,
                 max_hz=22050.,
                 sr=44100.,
                 smooth_l2mel=False):
        super(_l2melMatrixInitializer, self).__init__()
        self.min_hz = min_hz
        self.max_hz = max_hz
        self.sr = sr
        self.smooth_l2mel = smooth_l2mel

    def __call__(self, shape,
                 dtype=None):
        if dtype is None:
            dtype = tf.float32
        lower_edge_hz, higher_edge_hz = _get_lower_and_upper_edges_l2mel(self.min_hz, self.max_hz, shape[1])
        num_spectrogram_bins = shape[0]
        if self.smooth_l2mel:
            num_spectrogram_bins = min(shape(0) * 4, shape[1] * 8)

        l2melmat = tf.signal.linear_to_mel_weight_matrix(num_mel_bins=shape[1],
                                                         num_spectrogram_bins=num_spectrogram_bins,
                                                         lower_edge_hertz=lower_edge_hz,
                                                         upper_edge_hertz=higher_edge_hz,
                                                         sample_rate=self.sr,
                                                         dtype=dtype)
        if self.smooth_l2mel:
            l2melmat = tf.expand_dims(l2melmat, axis=-1)
            l2melmat = tf.image.resize(l2melmat, shape, method='mitchellcubic', antialias=True)
            l2melmat = tf.squeeze(l2melmat, axis=-1)

        return l2melmat

    def get_config(self):
        config = {'min_hz': self.min_hz,
                  'max_hz': self.max_hz,
                  'sr': self.sr,
                  'smooth_l2mel': self.smooth_l2mel}
        base_config = super(_l2melMatrixInitializer, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class STFT2MelspectrogramLayer(DynastesBaseLayer):

    def __init__(self,
                 n_mels,
                 min_hz=40.,
                 max_hz=20000.,
                 sr=44100.,
                 trainable=False,
                 smooth_l2mel=False,
                 ifreq=False,
                 return_phase=False,
                 **kwargs):
        super(STFT2MelspectrogramLayer, self).__init__(trainable=trainable, **kwargs)
        self.n_mels = n_mels
        self.min_hz = min_hz
        self.max_hz = max_hz
        self.sr = sr
        self.smooth_l2mel = smooth_l2mel
        self.ifreq = ifreq
        self.return_phase = return_phase
        self.inverted = False

    def build(self, input_shape):
        n_bins = input_shape[-2]
        l2mel_shape = [n_bins, self.n_mels]
        self.add_weight('l2mel', shape=l2mel_shape,
                        trainable=self.trainable,
                        initializer=_l2melMatrixInitializer(min_hz=self.min_hz,
                                                            max_hz=self.max_hz,
                                                            sr=self.sr,
                                                            smooth_l2mel=self.smooth_l2mel))

    def call(self, inputs, training=None, invert=False, **kwargs):
        if invert or self.inverted:
            stfts = spectral_ops.melspecgrams_to_stfts(melspecgrams=inputs,
                                                       mel2l=tf.transpose(self.get_weight('l2mel', training=training)),
                                                       ifreq=self.ifreq and self.return_phase)
            return stfts
        mel_spectrograms = spectral_ops.stfts_to_melspecgrams(stfts=inputs,
                                                              l2mel=self.get_weight('l2mel', training=training),
                                                              ifreq=self.ifreq,
                                                              return_phase=self.return_phase)
        mel_spectrograms = tf.cast(mel_spectrograms, self.dtype)
        return mel_spectrograms

    def get_config(self):
        config = {'n_mels': self.n_mels,
                  'min_hz': self.min_hz,
                  'max_hz': self.max_hz,
                  'sr': self.sr,
                  'smooth_l2mel': self.smooth_l2mel,
                  'ifreq': self.ifreq,
                  'return_phase': self.return_phase}
        base_config = super(STFT2MelspectrogramLayer, self).get_config()
        return {**base_config, **config}

    def compute_output_signature(self, input_signature, invert=False):
        if invert:
            self.inverted = True
            output_signature = super(STFT2MelspectrogramLayer, self).compute_output_signature(input_signature)
            self.inverted = False
            return output_signature
        else:
            input_shape = input_signature.shape.as_list()
            n_channels = input_shape[3]
            if self.return_phase:
                n_channels *= 2
            output_shape = tf.TensorShape([input_shape[0], input_shape[1], self.n_mels, n_channels])
            return tf.TensorSpec(shape=output_shape, dtype=self.dtype)

    def compute_output_shape(self, input_shape, invert=False):
        if invert:
            self.inverted = True
            output_shape = super(STFT2MelspectrogramLayer, self).compute_output_shape(input_shape)
            self.inverted = False
            return output_shape
        n_channels = input_shape[3]
        if self.return_phase:
            n_channels *= 2
        return tf.TensorShape([input_shape[0], input_shape[1], self.n_mels, n_channels])



from dynastes import activations
from dynastes import blocks
from dynastes import core
from dynastes import helpers
from dynastes import layers
from dynastes import models
from dynastes import ops
from dynastes import probability
from dynastes import regularizers
from dynastes import util
from dynastes import weight_normalizers
from dynastes.core import backend
from dynastes.probability import bijectors, bijector_partials, pseudoblocksparse_bijectors

__version__ = "0.8.2"



from .attention_layers import LocalizedAttentionLayer1D
from .attention_layers import LocalizedAttentionLayer2D
from .base_layers import ActivatedKernelBiasBaseLayer, _WscaleInitializer, DynastesDense
from .conditioning_layers import FeaturewiseLinearModulation
from .convolutional_layers import Downsampling1D
from .convolutional_layers import Downsampling2D
from .convolutional_layers import DynastesConv1D
from .convolutional_layers import DynastesConv1DTranspose
from .convolutional_layers import DynastesConv2D
from .convolutional_layers import DynastesConv2DTranspose
from .convolutional_layers import DynastesConv3D
from .convolutional_layers import DynastesDepthwiseConv1D
from .convolutional_layers import DynastesDepthwiseConv2D
from .convolutional_layers import Upsampling1D
from .convolutional_layers import Upsampling2D
from .normalization_layers import AdaptiveGroupNormalization
from .normalization_layers import AdaptiveInstanceNormalization
from .normalization_layers import AdaptiveLayerInstanceNormalization
from .normalization_layers import AdaptiveLayerNormalization
from .normalization_layers import AdaptiveMultiNormalization
from .normalization_layers import InstanceNormalization
from .normalization_layers import InstanceNormalization1D
from .normalization_layers import InstanceNormalization2D
from .normalization_layers import MultiNormalization
from .normalization_layers import PoolNormalization2D
from .random_layers import StatelessRandomNormalLike
from .spectral_layers import STFT2MelspectrogramLayer
from .spectral_layers import Wave2STFTLayer
from .t2t_attention_layers import AddTimingSignalLayer1D
from .t2t_attention_layers import Attention1D
from .t2t_attention_layers import Attention2D
from .t2t_attention_layers import LshGatingLayer
from .t2t_attention_layers import PseudoBlockSparseAttention1D
from .time_delay_layers import DepthGroupwiseTimeDelayLayer1D
from .time_delay_layers import DepthGroupwiseTimeDelayLayerFake2D
from .time_delay_layers import TimeDelayLayer1D
from .time_delay_layers import TimeDelayLayerFake2D



import numpy as np
import tensorflow as tf
import dynastes.layers as dynl
from dynastes.layers.convolutional_layers import DynastesConv1DTranspose, DynastesDepthwiseConv1D, \
    DynastesConv2DTranspose, DynastesConv2D, DynastesDepthwiseConv2D, DynastesConv1D, Upsampling1D, Upsampling2D
from dynastes.util.test_utils import layer_test

to_tensor = tf.convert_to_tensor
normal = np.random.normal


class DynastesConv1DTest(tf.test.TestCase):

    def test_masking(self):
        layer = DynastesConv1D(32, kernel_size=3, strides=2, padding='same')

        ts = to_tensor(normal(size=(8, 16, 32))
                       .astype(np.float32))

        mask_len = 16 // 6
        mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
        mask = tf.expand_dims(mask, axis=0)
        mask = tf.tile(mask, [8, 1])
        layer(ts, mask=mask)
        layer.compute_mask(ts, mask=mask)

        @tf.function
        def graph_test_fn(x, mask):
            layer(x, mask=mask)
            layer.compute_mask(x, mask=mask)

        graph_test_fn(x=ts, mask=mask)

    def test_simple(self):
        layer_test(
            DynastesConv1D, kwargs={'filters': 3, 'kernel_size': 3, 'activation': 'mish'}, input_shape=(5, 32, 3))

    def test_specnorm(self):
        layer_test(
            DynastesConv1D, kwargs={'filters': 3, 'kernel_size': 3,
                                    'kernel_normalizer': 'spectral',
                                    'use_wscale': True,
                                    'kernel_regularizer': 'orthogonal'}, input_shape=(5, 32, 3))


class DynastesConv2DTest(tf.test.TestCase):
    def test_simple(self):
        layer_test(
            DynastesConv2D, kwargs={'filters': 3, 'kernel_size': (3, 3), 'activation': 'swish'},
            input_shape=(4, 16, 16, 3))

    def test_masking(self):
        layer = DynastesConv2D(32, kernel_size=3, strides=2, padding='same')

        ts = to_tensor(normal(size=(8, 16, 16, 32))
                       .astype(np.float32))

        mask_len = 16 // 8
        mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
        mask = tf.expand_dims(mask, axis=0)
        mask = tf.tile(mask, [8, 1])
        mask = tf.expand_dims(mask, axis=1)
        mask = tf.tile(mask, [1, 16, 1])
        layer(ts, mask=mask)
        layer.compute_mask(ts, mask=mask)

        @tf.function
        def graph_test_fn(x, mask):
            layer(x, mask=mask)
            layer.compute_mask(x, mask=mask)

        graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        layer_test(
            DynastesConv2D, kwargs={'filters': 7, 'kernel_size': (3, 3),
                                    'kernel_normalizer': 'spectral',
                                    'kernel_regularizer': 'orthogonal'}, input_shape=(4, 16, 16, 5))


class DynastesConv1DTransposeTest(tf.test.TestCase):
    def test_simple(self):
        layer_test(
            DynastesConv1DTranspose,
            kwargs={'filters': 3, 'kernel_size': 3, 'strides': 2, 'padding': 'same'},
            input_shape=(None, 16, 3),
            expected_output_shape=(None, 32, 3)
        )

    def test_masking(self):
        layer = DynastesConv1DTranspose(32, kernel_size=3, strides=2, padding='same')

        ts = to_tensor(normal(size=(8, 16, 32))
                       .astype(np.float32))

        mask_len = 16 // 6
        mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
        mask = tf.expand_dims(mask, axis=0)
        mask = tf.tile(mask, [8, 1])
        layer(ts, mask=mask)
        layer.compute_mask(ts, mask=mask)

        @tf.function
        def graph_test_fn(x, mask):
            layer(x, mask=mask)
            layer.compute_mask(x, mask=mask)

        graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        layer_test(
            DynastesConv1DTranspose,
            kwargs={'filters': 7, 'kernel_size': 3, 'strides': 2, 'padding': 'same',
                    'kernel_normalizer': 'spectral',
                    'kernel_regularizer': 'orthogonal'},
            input_shape=(None, 16, 5),
            expected_output_shape=(None, 32, 7)
        )


class DynastesConv2DTransposeTest(tf.test.TestCase):
    def test_simple(self):
        layer_test(
            DynastesConv2DTranspose,
            kwargs={'filters': 3, 'kernel_size': (3, 3), 'strides': (2, 2), 'padding': 'same'},
            input_shape=(None, 16, 16, 3),
            expected_output_shape=(None, 32, 32, 3)
        )

    def test_masking(self):
        layer = DynastesConv2DTranspose(32, kernel_size=3, strides=2, padding='same')

        ts = to_tensor(normal(size=(8, 16, 16, 32))
                       .astype(np.float32))

        mask_len = 16 // 8
        mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
        mask = tf.expand_dims(mask, axis=0)
        mask = tf.tile(mask, [8, 1])
        mask = tf.expand_dims(mask, axis=1)
        mask = tf.tile(mask, [1, 16, 1])
        layer(ts, mask=mask)
        layer.compute_mask(ts, mask=mask)

        @tf.function
        def graph_test_fn(x, mask):
            layer(x, mask=mask)
            layer.compute_mask(x, mask=mask)

        graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        layer_test(
            DynastesConv2DTranspose,
            kwargs={'filters': 7, 'kernel_size': (3, 3), 'strides': (2, 2), 'padding': 'same',
                    'kernel_normalizer': 'spectral',
                    'kernel_regularizer': 'orthogonal'},
            input_shape=(None, 16, 16, 5),
            expected_output_shape=(None, 32, 32, 7)
        )


class DynastesDepthwiseConv1DTest(tf.test.TestCase):
    def test_simple(self):
        layer_test(
            DynastesDepthwiseConv1D, kwargs={'kernel_size': 3}, input_shape=(5, 32, 3))

    def test_masking(self):
        layer = DynastesDepthwiseConv1D(kernel_size=3, strides=2, padding='same')

        ts = to_tensor(normal(size=(8, 16, 32))
                       .astype(np.float32))

        mask_len = 16 // 6
        mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
        mask = tf.expand_dims(mask, axis=0)
        mask = tf.tile(mask, [8, 1])
        layer(ts, mask=mask)
        layer.compute_mask(ts, mask=mask)

        @tf.function
        def graph_test_fn(x, mask):
            layer(x, mask=mask)
            layer.compute_mask(x, mask=mask)

        graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        layer_test(
            DynastesDepthwiseConv1D, kwargs={'kernel_size': 3,
                                             'kernel_normalizer': 'spectral',
                                             'kernel_regularizer': 'orthogonal'}, input_shape=(5, 32, 3))


class Upsampling2DTest(tf.test.TestCase):
    def test_simple(self):
        layer_test(
            Upsampling2D,
            kwargs={'strides': (2, 2)},
            input_shape=(None, 16, 16, 3),
            expected_output_shape=(None, 32, 32, 3)
        )

    def test_masking(self):
        layer = Upsampling2D(strides=2)

        ts = to_tensor(normal(size=(8, 16, 16, 32))
                       .astype(np.float32))

        mask_len = 16 // 8
        mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
        mask = tf.expand_dims(mask, axis=0)
        mask = tf.tile(mask, [8, 1])
        mask = tf.expand_dims(mask, axis=1)
        mask = tf.tile(mask, [1, 16, 1])
        layer(ts, mask=mask)
        layer.compute_mask(ts, mask=mask)

        @tf.function
        def graph_test_fn(x, mask):
            layer(x, mask=mask)
            layer.compute_mask(x, mask=mask)

        graph_test_fn(x=ts, mask=mask)


class Upsampling1DTest(tf.test.TestCase):
    def test_simple(self):
        layer_test(
            Upsampling1D, kwargs={'strides': 2}, input_shape=(None, 32, 3), expected_output_shape=(None, 64, 3))

    def test_masking(self):
        layer = Upsampling1D(strides=2)

        ts = to_tensor(normal(size=(8, 16, 32))
                       .astype(np.float32))

        mask_len = 16 // 6
        mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
        mask = tf.expand_dims(mask, axis=0)
        mask = tf.tile(mask, [8, 1])
        layer(ts, mask=mask)
        layer.compute_mask(ts, mask=mask)

        @tf.function
        def graph_test_fn(x, mask):
            layer(x, mask=mask)
            layer.compute_mask(x, mask=mask)

        graph_test_fn(x=ts, mask=mask)


class DynastesDepthwiseConv2DTest(tf.test.TestCase):
    def test_simple(self):
        layer_test(
            DynastesDepthwiseConv2D, kwargs={'kernel_size': (3, 3)}, input_shape=(4, 16, 16, 3))

    def test_masking(self):
        layer = DynastesDepthwiseConv2D(kernel_size=3, strides=2, padding='same')

        ts = to_tensor(normal(size=(8, 16, 16, 32))
                       .astype(np.float32))

        mask_len = 16 // 6
        mask = to_tensor(([True] * (16 - mask_len)) + ([False] * (mask_len)))
        mask = tf.expand_dims(mask, axis=0)
        mask = tf.tile(mask, [8, 1])
        mask = tf.expand_dims(mask, axis=1)
        mask = tf.tile(mask, [1, 16, 1])
        layer(ts, mask=mask)
        layer.compute_mask(ts, mask=mask)

        @tf.function
        def graph_test_fn(x, mask):
            layer(x, mask=mask)
            layer.compute_mask(x, mask=mask)

        graph_test_fn(x=ts, mask=mask)

    def test_specnorm(self):
        layer_test(
            DynastesDepthwiseConv2D, kwargs={'kernel_size': (3, 3),
                                             'kernel_normalizer': 'spectral'}, input_shape=(4, 16, 16, 5))



import tensorflow as tf


def custom_gradient_back_fn(back_f):
    def _recompute_grad_custom_back_fn(f):
        """An eager-compatible version of recompute_grad.
        For f(*args, **kwargs), this supports gradients with respect to args, or to
        gradients with respect to any variables residing in the kwarg 'variables'.
        Note that for keras layer and model objects, this is handled automatically.
        Warning: If `f` was originally a tf.keras Model or Layer object, `g` will not
        be able to access the member variables of that object, because `g` returns
        through the wrapper function `inner`.  When recomputing gradients through
        objects that inherit from keras, we suggest keeping a reference to the
        underlying object around for the purpose of accessing these variables.
        Args:
          f: function `f(*x)` that returns a `Tensor` or sequence of `Tensor` outputs.
        Returns:
         A function `g` that wraps `f`, but which recomputes `f` on the backwards
         pass of a gradient call.
        """

        # TODO(cdfreeman) Add is_recomputing functionality from graph mode version

        @tf.custom_gradient
        def inner(*args, **kwargs):
            """Inner function closure for calculating gradients."""
            result = f(*args, **kwargs)

            def grad(dresult, variables=None):
                """Gradient function calculation for inner function."""
                with tf.GradientTape() as t:
                    t.watch(args)
                    if variables is not None:
                        t.watch(variables)
                    with tf.control_dependencies([dresult]):
                        result = back_f(*args, **kwargs)
                kw_vars = []
                if variables is not None:
                    kw_vars = list(variables)
                grads = t.gradient(
                    result, list(args) + kw_vars, output_gradients=[dresult])
                return grads[:len(args)], grads[len(args):]

            return result, grad

        return inner

    return _recompute_grad_custom_back_fn



import tensorflow as tf

from dynastes.layers import DynastesDense
from dynastes.util.test_utils import layer_test


class DynastesDenseTest1D(tf.test.TestCase):
    def test_simple(self):
        layer_test(
            DynastesDense, kwargs={'units': 12}, input_shape=(5, 32, 3))

    def test_specnorm(self):
        layer_test(
            DynastesDense, kwargs={'units': 12,
                                   'kernel_normalizer': 'spectral',
                                   'use_wscale': True,
                                   'kernel_regularizer': 'orthogonal'}, input_shape=(5, 32, 3))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from tensorflow.python.keras import activations as tf_activations
from tensorflow.python.keras.activations import serialize as _serialize

tfaa = None
try:
    import tensorflow_addons.activations as tfaa
except:
    print('WARNING! TensorFlow Addons are missing!')
    tfal = None


@tf.keras.utils.register_keras_serializable(package='Dynastes', name='swish')
def swish(x, beta=1):
    return x * tf.math.sigmoid(beta * x)


mish = None
rrelu = None

if tfaa is not None:
    mish = tfaa.mish
    rrelu = tfaa.rrelu


def get(activation):
    if type(activation) == str:
        if activation == 'swish':
            return swish
        elif tfaa is not None:
            if activation == 'mish':
                return mish
    return tf_activations.get(activation)


def serialize(activation):
    return _serialize(activation)



import timeit

import numpy as np
import tensorflow as tf
from tensorflow.keras import layers as tfkl
from tensorflow.python.framework import test_util

from dynastes.models.growing_gan_models import GrowingGanGenerator, GrowingGanClassifier


class Simple2DGrowingGanGenerator(GrowingGanGenerator):

    def __init__(self,
                 n_lods,
                 **kwargs):
        super(Simple2DGrowingGanGenerator, self).__init__(n_lods=n_lods, **kwargs)
        self.gan_layers = [tfkl.UpSampling2D(interpolation='bilinear') for _ in range(n_lods)]
        self.dom_layers = [tfkl.Activation('linear') for _ in range(n_lods + 1)]

    def interpolate_domain(self, x, y, interp):
        return x + (y - x) * interp

    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:

        if lod > (self.n_lods - 1):
            # Return latent
            return tfkl.Activation('linear')
        return self.gan_layers[(self.n_lods - 1) - lod]

    def get_to_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.dom_layers[(self.n_lods - 1) - lod]

    def get_upscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        if output_lod == input_lod:
            return tfkl.Activation('linear')
        levels = [2] * (output_lod - input_lod)
        if len(levels) == 0:
            return tfkl.Activation('linear')
        scale = np.cumprod(levels)[-1]
        return tfkl.UpSampling2D(size=(scale, scale))

    def get_conform_to_output_layer_by_lod(self, input_lod) -> tfkl.Layer:
        return self.get_upscale_domain_layer_by_lod(input_lod, self.n_lods - 1)


class GrowingGanGeneratorTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        in_z_hw = 2
        n_lods = 4
        z = np.random.random(size=(1, in_z_hw, in_z_hw, 1)).astype(np.float32)
        gen = Simple2DGrowingGanGenerator(n_lods=n_lods)
        y = gen(z, lod_in=2.)
        ex_hw = in_z_hw * (2 ** (n_lods))
        self.assertShapeEqual(np.ones(shape=(1, ex_hw, ex_hw, 1)), y)


class Real2DGrowingGanGenerator(GrowingGanGenerator):

    def __init__(self,
                 n_lods,
                 **kwargs):
        super(Real2DGrowingGanGenerator, self).__init__(n_lods=n_lods, **kwargs)
        self.gan_layers = [tfkl.Conv2DTranspose(2 ** (n_lods - i), kernel_size=(3, 3), strides=(2, 2), padding='same')
                           for i in range(n_lods)]
        self.dom_layers = [tfkl.Conv2D(3, kernel_size=(1, 1), strides=(1, 1), padding='same') for _ in
                           range(n_lods + 1)]

    def interpolate_domain(self, x, y, interp):
        return x + (y - x) * interp

    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        if lod > (self.n_lods - 1):
            return tfkl.Activation('linear')
        return self.gan_layers[(self.n_lods - 1) - lod]

    def get_to_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.dom_layers[(self.n_lods - 1) - lod]

    def get_upscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        if output_lod == input_lod:
            return tfkl.Activation('linear')
        levels = [2] * (output_lod - input_lod)
        if len(levels) == 0:
            return tfkl.Activation('linear')
        scale = np.cumprod(levels)[-1]
        return tfkl.UpSampling2D(size=(scale, scale))

    def get_conform_to_output_layer_by_lod(self, input_lod) -> tfkl.Layer:
        return self.get_upscale_domain_layer_by_lod(input_lod, self.n_lods - 1)


class GrowingGanGeneratorTestReal(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        in_z_hw = 2
        n_lods = 2
        z = np.random.random(size=(1, in_z_hw, in_z_hw, 2 ** (n_lods))).astype(np.float32)
        gen = Real2DGrowingGanGenerator(n_lods)
        y = gen(z, lod_in=0.5)
        ex_hw = in_z_hw * (2 ** (n_lods))
        self.assertShapeEqual(np.ones(shape=(1, ex_hw, ex_hw, 3)), y)


class Complex2DGrowingGanGenerator(GrowingGanGenerator):

    def __init__(self,
                 strides=[1, 2, 3],
                 **kwargs):
        n_lods = len(strides)
        super(Complex2DGrowingGanGenerator, self).__init__(n_lods=n_lods, **kwargs)

        self.strides = strides
        channels = [2 ** (n_lods - i) for i in range(n_lods)]
        self.gan_layers = [tfkl.Conv2DTranspose(ch, kernel_size=(3, 3), strides=(s, s), padding='same') for ch, s in
                           zip(channels, self.strides)]
        self.dom_layers = [tfkl.Conv2D(3, kernel_size=(1, 1), strides=(1, 1), padding='same') for _ in
                           range(self.n_lods + 1)]

    def interpolate_domain(self, x, y, interp):
        return x + (y - x) * interp

    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        if lod > (self.n_lods - 1):
            return tfkl.Activation('linear')
        return self.gan_layers[(self.n_lods - 1) - lod]

    def get_to_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.dom_layers[(self.n_lods - 1) - lod]

    def get_upscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        if output_lod == input_lod:
            return tfkl.Activation('linear')
        input_idx = max((self.n_lods - 1) - output_lod, 0)
        output_idx = max((self.n_lods - 1) - input_lod, 0)
        strides = (self.strides)[input_idx: output_idx]
        if len(strides) == 0:
            return tfkl.Activation('linear')
        scale = np.cumprod(strides)[-1]
        return tfkl.UpSampling2D(size=(scale, scale))

    def get_conform_to_output_layer_by_lod(self, input_lod) -> tfkl.Layer:
        return self.get_upscale_domain_layer_by_lod(input_lod, self.n_lods)


class GrowingGanGeneratorTestComplex(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        print('GrowingGanGeneratorTestComplex - test_simple')
        in_z_hw = 1
        strides = [7, 2, 3, 1, 3, 5, 1]
        z = np.random.random(size=(1, in_z_hw, in_z_hw, 2 ** len(strides) + 1)).astype(np.float32)
        gen = Complex2DGrowingGanGenerator(strides)
        for i in range(len(strides) * 2):
            y = gen(z, lod_in=i / 2)
            ex_hw = in_z_hw * np.cumprod(strides)[-1]
            self.assertShapeEqual(np.ones(shape=(1, ex_hw, ex_hw, 3)), y)
        max_lod = float(len(strides) + 1)

        # Test timing
        def lod_0():
            y = gen(z, lod_in=0.)

        def lod_max():
            y = gen(z, lod_in=max_lod)

        time_l0 = timeit.timeit(lod_0, number=2)
        time_l7 = timeit.timeit(lod_max, number=2)
        print('LOD time diff', (time_l7 / time_l0))


class Simple2DGrowingGanClassifier(GrowingGanClassifier):

    def interpolate_domain(self, x, y, interp):
        return x + (y - x) * interp

    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        return tfkl.AveragePooling2D((2, 2), strides=(2, 2))

    def get_from_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        return tfkl.Activation('linear')

    def get_downscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        if output_lod == input_lod:
            return tfkl.Activation('linear')
        levels = [2] * (input_lod - output_lod)
        if len(levels) == 0:
            return tfkl.Activation('linear')
        scale = np.cumprod(levels)[-1]
        return tfkl.AveragePooling2D((scale, scale), strides=(scale, scale))

    def get_input_transform_layer_by_lod(self, lod) -> tfkl.Layer:

        return self.get_downscale_domain_layer_by_lod(self.n_lods, lod)


class GrowingGanClassifierTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        in_hw = 32
        n_lods = 3
        z = np.random.random(size=(1, in_hw, in_hw, 1)).astype(np.float32)
        cls = Simple2DGrowingGanClassifier(n_lods=n_lods)
        y = cls(z, lod_in=1.)
        ex_hw = in_hw // (2 ** (n_lods))
        self.assertShapeEqual(np.ones(shape=(1, ex_hw, ex_hw, 1)), y)


class GrowingEndToEndTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        n_lods = 3
        z = np.random.random(size=(1, 64, 48, 1)).astype(np.float32)
        z = tf.convert_to_tensor(z)
        gen = Simple2DGrowingGanGenerator(n_lods=n_lods)
        cls = Simple2DGrowingGanClassifier(n_lods=n_lods)
        _z = z
        for i in range(int(n_lods * 2.)):
            _z = gen(cls(_z, lod_in=i / 2.), lod_in=i / 2.)
        self.assertLess(tf.reduce_mean(tf.abs(z - _z)).numpy(), 0.26)


class Real2DGrowingGanClassifier(GrowingGanClassifier):

    def __init__(self, n_lods=3,
                 **kwargs):
        super(Real2DGrowingGanClassifier, self).__init__(n_lods=n_lods, **kwargs)
        self.gan_layers = [tfkl.Conv2D(2 ** (i + 1), kernel_size=(3, 3), strides=(2, 2), padding='same') for i in
                           range(0, n_lods)]
        self.from_dom_layers = [tfkl.Conv2D(2 ** (i), kernel_size=(1, 1), strides=(1, 1), padding='same') for i in
                                range(0, n_lods + 1)]

    def interpolate_domain(self, x, y, interp):
        return x + (y - x) * interp

    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        if lod < 0:
            raise ArithmeticError
        return self.gan_layers[(self.n_lods - lod)]

    def get_from_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.from_dom_layers[(self.n_lods - lod)]

    def get_downscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        if output_lod == input_lod:
            return tfkl.Activation('linear')
        levels = [2] * (input_lod - output_lod)
        if len(levels) == 0:
            return tfkl.Activation('linear')
        scale = np.cumprod(levels)[-1]
        if scale == 1:
            return tfkl.Activation('linear')
        return tfkl.AveragePooling2D((scale, scale), strides=(scale, scale), padding='same')

    def get_input_transform_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.get_downscale_domain_layer_by_lod(self.n_lods, lod)


class GrowingGanClassifierTestReal(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        in_hw = 32
        n_lods = 4
        z = np.random.random(size=(1, in_hw, in_hw, 1)).astype(np.float32)
        cls = Real2DGrowingGanClassifier(n_lods=n_lods)
        y = cls(z, lod_in=3.)
        ex_hw = in_hw // (2 ** (n_lods))
        self.assertShapeEqual(np.ones(shape=(1, ex_hw, ex_hw, 2 ** n_lods)), y)


class ComplexGanClassifier(GrowingGanClassifier):

    def __init__(self,
                 strides=[2, 3, 4, 2, 5],
                 **kwargs):
        n_lods = len(strides)
        super(ComplexGanClassifier, self).__init__(n_lods=n_lods, **kwargs)
        self.strides = strides
        channels = [2 ** (i + 1) for i in range(len(self.strides) + 1)]
        # channels = [3, 4, 4, 8, 16, 32]
        self.gan_layers = []
        self.dom_layers = []
        for i, (s, ch) in enumerate(zip(self.strides, channels[1:])):
            self.gan_layers.append(tfkl.Conv2D(ch, kernel_size=(3, 3), strides=(s, s), padding='same'))
        for ch in channels:
            self.dom_layers.append(tfkl.Conv2D(ch, kernel_size=(1, 1), strides=(1, 1), padding='same'))

    def interpolate_domain(self, x, y, interp):
        return x + (y - x) * interp

    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        if lod < 0:
            raise ArithmeticError
        return self.gan_layers[(self.n_lods - lod)]

    def get_from_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.dom_layers[(self.n_lods - lod)]

    def get_downscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        if output_lod == input_lod:
            return tfkl.Activation('linear')
        strides = self.strides[self.n_lods - input_lod: self.n_lods - output_lod]
        if len(strides) == 0:
            return tfkl.Activation('linear')
        scale = np.cumprod(strides)[-1]
        if scale == 1:
            return tfkl.Activation('linear')
        return tfkl.AveragePooling2D((scale, scale), strides=(scale, scale), padding='same')

    def get_input_transform_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.get_downscale_domain_layer_by_lod(self.n_lods, lod)


class GrowingGanClassifierTestComplex(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        strides = [1, 3, 7, 5, 3]
        in_base = 1
        in_hw = in_base * np.cumprod(strides)[-1]
        n_lods = len(strides)
        z = np.random.random(size=(1, in_hw, in_hw, 1)).astype(np.float32)
        cls = ComplexGanClassifier(strides)
        y = cls(z, lod_in=3.)
        ex_size_hw = in_hw // np.cumprod(strides)[-1]
        self.assertShapeEqual(np.ones(shape=(1, ex_size_hw, ex_size_hw, 2 ** (n_lods + 1))), y)



import os

import tensorflow as tf
from tensorflow.python.framework import test_util

from dynastes.ops import spectral_ops


class SpectralOpsTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        print('### Running spectral ops tests ###')

        n_fft = 1024
        fft_bins = n_fft // 2
        hop_length = 256
        pad = 8192 * 4
        mel_upsampling = 1
        mel_mul = 2
        dir_path = os.path.dirname(os.path.realpath(__file__))

        raw_audio = tf.io.read_file(dir_path + '/data/test_wav.wav')
        waves = tf.audio.decode_wav(raw_audio)[0][:20480] * 0.25
        waves = tf.expand_dims(waves, 0)
        waves = tf.expand_dims(waves, -1)
        waves = tf.keras.layers.UpSampling2D((2, 1), interpolation='bilinear')(waves)
        waves = tf.squeeze(waves, axis=-1)

        stfts = spectral_ops.waves_to_stfts(waves, n_fft=n_fft, hop_length=hop_length, pad_l=pad, pad_r=pad,
                                            discard_dc=True)
        stfts_mags = tf.maximum(tf.minimum(tf.abs(stfts), 0.5), -0.5)
        stfts_angles = tf.math.angle(stfts)
        stfts = spectral_ops.polar2rect(stfts_mags, stfts_angles)

        waves_recon = spectral_ops.stfts_to_waves(stfts, n_fft=n_fft, hop_length=hop_length, pad_l=pad, pad_r=pad,
                                                  discard_dc=True)
        wave_err = tf.abs(tf.reduce_mean(tf.abs(waves) - tf.abs(waves_recon))).numpy()
        wave_dc_err = tf.abs(tf.reduce_mean(waves - waves_recon)).numpy()
        print('wave_err', wave_err, 'wave_dc_err', wave_dc_err)
        assert (wave_err < 0.03)
        assert (wave_dc_err < 0.001)
        waves_recon_stft = spectral_ops.waves_to_stfts(waves_recon, n_fft=n_fft, hop_length=hop_length, pad_l=pad,
                                                       pad_r=pad, discard_dc=True)
        stft_err = tf.abs(tf.reduce_mean(tf.abs(stfts) - tf.abs(waves_recon_stft))).numpy()
        stft_mean_err = tf.abs(tf.reduce_mean(stfts - waves_recon_stft)).numpy()
        print('stft_err', stft_err, 'stft_mean_err', stft_mean_err)
        assert (stft_err < 0.005)
        assert (stft_mean_err < 1e-4)
        l2mel = tf.signal.linear_to_mel_weight_matrix(num_spectrogram_bins=fft_bins,
                                                      sample_rate=44100,
                                                      lower_edge_hertz=80.,
                                                      upper_edge_hertz=44100/2,
                                                      num_mel_bins=(fft_bins * mel_upsampling) * mel_mul)

        def process_l2mel_mat(l2mel, iterations=200):

            def iterate_l2mel(l2mel):
                x = tf.expand_dims(l2mel, axis=-1)
                x = tf.keras.layers.AveragePooling2D((1,2), strides=(1,2))(x)
                x = tf.keras.layers.UpSampling2D((1,2), interpolation='bilinear')(x)
                x = tf.squeeze(x, axis=-1)
                mask = tf.cast(l2mel < 1e-16, x.dtype)
                x = x * mask + l2mel * (1-mask)
                x = (x * 0.5) + (l2mel * 0.5)
                xs = tf.reduce_sum(x, axis=-1, keepdims=True)
                ls = tf.reduce_sum(l2mel, axis=-1, keepdims=True)
                return x * (ls/(xs + 1e-16))

            l2mel = tf.expand_dims(l2mel, axis=0)
            for i in range(iterations):
                l2mel = iterate_l2mel(l2mel)
                l2mel = tf.transpose(iterate_l2mel(tf.transpose(l2mel, [0,2,1])), [0,2,1])

            return tf.squeeze(l2mel, axis=0)

        l2mel = process_l2mel_mat(l2mel)
        print(l2mel.numpy())
        print(l2mel.shape)
        # Mel transform
        melspecgrams = spectral_ops.stfts_to_melspecgrams(stfts, l2mel=l2mel)

        # Using mel-upsampling yields much lower reconstruction error!
        if mel_upsampling > 1:
            melspecgrams = tf.keras.layers.AveragePooling2D((1, mel_upsampling), strides=(1, mel_upsampling),
                                                            padding='same', data_format='channels_last', dtype='float64')(melspecgrams)
            # Simulate output downsampling
            melspecgrams = tf.keras.layers.UpSampling2D((1, mel_upsampling), data_format='channels_last',
                                                        interpolation='nearest', dtype='float64')(melspecgrams)

        stfts_mel_recon = spectral_ops.melspecgrams_to_stfts(melspecgrams, mel2l=tf.transpose(l2mel))
        waves_mel_recon = spectral_ops.stfts_to_waves(stfts_mel_recon, n_fft=n_fft, hop_length=hop_length,
                                                      discard_dc=True, pad_l=pad, pad_r=pad)
        melspec_wav_recon_err = tf.abs(tf.reduce_mean(waves - waves_mel_recon)).numpy()
        print('melspec_wav_recon_err', melspec_wav_recon_err)
        assert (melspec_wav_recon_err < 0.0002)

        melspecgram_stfts_mel_recon = spectral_ops.stfts_to_melspecgrams(stfts_mel_recon, l2mel=l2mel)

        melspecgrams_err = tf.abs(tf.reduce_mean(tf.abs(melspecgrams) - tf.abs(melspecgram_stfts_mel_recon))).numpy()
        melspecgrams_mean_err = tf.abs(tf.reduce_mean(melspecgrams - melspecgram_stfts_mel_recon)).numpy()
        print('melspecgrams_err', melspecgrams_err, 'melspecgrams_mean_err', melspecgrams_mean_err)
        assert (melspecgrams_err < 0.02)
        assert (melspecgrams_mean_err < 0.021)

        stfts_mel_recon_mel_recon = spectral_ops.melspecgrams_to_stfts(melspecgram_stfts_mel_recon,
                                                                       mel2l=tf.transpose(l2mel))
        mel_recon_err = tf.abs(tf.reduce_mean(tf.abs(stfts_mel_recon) - tf.abs(stfts_mel_recon_mel_recon))).numpy()
        print('mel_recon_err', mel_recon_err)
        assert (mel_recon_err < 0.0025)
        waves_mel_recon = spectral_ops.stfts_to_waves(stfts_mel_recon, n_fft=n_fft, hop_length=hop_length, pad_l=pad,
                                                      pad_r=pad, discard_dc=True)
        waves_mel_recon_mel_recon = spectral_ops.stfts_to_waves(stfts_mel_recon_mel_recon, n_fft=n_fft,
                                                                hop_length=hop_length, pad_l=pad, pad_r=pad,
                                                                discard_dc=True)
        assert (tf.abs(tf.reduce_mean(tf.abs(waves_mel_recon) - tf.abs(waves_mel_recon_mel_recon))).numpy() < 0.05)




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading
import tensorflow as tf
import numpy as np
from tensorflow.python import keras
from tensorflow.python.eager import context
from tensorflow.python.framework import tensor_shape
from tensorflow.python.framework import tensor_spec
from tensorflow.python.framework import test_util
from tensorflow.python.util import tf_inspect

_thread_local_data = threading.local()
_thread_local_data.model_type = None
_thread_local_data.run_eagerly = None
_thread_local_data.experimental_run_tf_function = None
_thread_local_data.saved_model_format = None


def should_run_eagerly():
    """Returns whether the models we are testing should be run eagerly."""
    if _thread_local_data.run_eagerly is None:
        raise ValueError('Cannot call `should_run_eagerly()` outside of a '
                         '`run_eagerly_scope()` or `run_all_keras_modes` '
                         'decorator.')

    return _thread_local_data.run_eagerly and context.executing_eagerly()


@test_util.use_deterministic_cudnn
def layer_test(layer_cls, kwargs=None, input_shape=None, input_dtype=None,
               input_data=None, expected_output=None,
               expected_output_dtype=None, expected_output_shape=None,
               validate_training=True, adapt_data=None):
    """Test routine for a layer with a single input and single output.

    Arguments:
      layer_cls: Layer class object.
      kwargs: Optional dictionary of keyword arguments for instantiating the
        layer.
      input_shape: Input shape tuple.
      input_dtype: Data type of the input data.
      input_data: Numpy array of input data.
      expected_output: Numpy array of the expected output.
      expected_output_dtype: Data type expected for the output.
      expected_output_shape: Shape tuple for the expected shape of the output.
      validate_training: Whether to attempt to validate training on this layer.
        This might be set to False for non-differentiable layers that output
        string or integer values.
      adapt_data: Optional data for an 'adapt' call. If None, adapt() will not
        be tested for this layer. This is only relevant for PreprocessingLayers.

    Returns:
      The output data (Numpy array) returned by the layer, for additional
      checks to be done by the calling code.

    Raises:
      ValueError: if `input_shape is None`.
    """
    if input_data is None:
        if input_shape is None:
            raise ValueError('input_shape is None')
        if not input_dtype:
            input_dtype = 'float32'
        input_data_shape = list(input_shape)
        for i, e in enumerate(input_data_shape):
            if e is None:
                input_data_shape[i] = np.random.randint(1, 4)
        input_data = 10 * np.random.random(input_data_shape)
        if input_dtype in [tf.complex64, tf.complex128]:
            input_data = tf.complex(input_data, tf.convert_to_tensor(0.0, dtype=input_data.dtype))
            input_data = tf.cast(input_data, input_dtype).numpy()
        else:
            if input_dtype == 'float':
                input_data -= 0.5
            input_data = input_data.astype(input_dtype)
    elif input_shape is None:
        input_shape = input_data.shape
    if input_dtype is None:
        input_dtype = input_data.dtype
    if expected_output_dtype is None:
        expected_output_dtype = input_dtype

    # instantiation
    kwargs = kwargs or {}
    layer = layer_cls(**kwargs)

    # Test adapt, if data was passed.
    if adapt_data is not None:
        layer.adapt(adapt_data)

    # test get_weights , set_weights at layer level
    weights = layer.get_weights()
    layer.set_weights(weights)

    # test and instantiation from weights
    if 'weights' in tf_inspect.getargspec(layer_cls.__init__):
        kwargs['weights'] = weights
        layer = layer_cls(**kwargs)

    # test in functional API
    x = keras.layers.Input(shape=input_shape[1:], dtype=input_dtype)
    y = layer(x)
    if keras.backend.dtype(y) != expected_output_dtype:
        raise AssertionError('When testing layer %s, for input %s, found output '
                             'dtype=%s but expected to find %s.\nFull kwargs: %s' %
                             (layer_cls.__name__,
                              x,
                              keras.backend.dtype(y),
                              expected_output_dtype,
                              kwargs))

    def assert_shapes_equal(expected, actual):
        """Asserts that the output shape from the layer matches the actual shape."""
        if len(expected) != len(actual):
            raise AssertionError(
                'When testing layer %s, for input %s, found output_shape='
                '%s but expected to find %s.\nFull kwargs: %s' %
                (layer_cls.__name__, x, actual, expected, kwargs))

        for expected_dim, actual_dim in zip(expected, actual):
            if isinstance(expected_dim, tensor_shape.Dimension):
                expected_dim = expected_dim.value
            if isinstance(actual_dim, tensor_shape.Dimension):
                actual_dim = actual_dim.value
            if expected_dim is not None and expected_dim != actual_dim:
                raise AssertionError(
                    'When testing layer %s, for input %s, found output_shape='
                    '%s but expected to find %s.\nFull kwargs: %s' %
                    (layer_cls.__name__, x, actual, expected, kwargs))

    if expected_output_shape is not None:
        assert_shapes_equal(tensor_shape.TensorShape(expected_output_shape),
                            y.shape)

    # check shape inference
    model = keras.models.Model(x, y)
    _comp_output_shape = layer.compute_output_shape(
        tensor_shape.TensorShape(input_shape))
    computed_output_shape = tuple(_comp_output_shape.as_list())
    computed_output_signature = layer.compute_output_signature(
        tensor_spec.TensorSpec(shape=input_shape, dtype=input_dtype))
    actual_output = model.predict(input_data)
    actual_output_shape = actual_output.shape
    assert_shapes_equal(computed_output_shape, actual_output_shape)
    assert_shapes_equal(computed_output_signature.shape, actual_output_shape)
    if computed_output_signature.dtype != actual_output.dtype:
        raise AssertionError(
            'When testing layer %s, for input %s, found output_dtype='
            '%s but expected to find %s.\nFull kwargs: %s' %
            (layer_cls.__name__, x, actual_output.dtype,
             computed_output_signature.dtype, kwargs))
    if expected_output is not None:
        np.testing.assert_allclose(actual_output, expected_output,
                                   rtol=1e-3, atol=1e-6)

    # test serialization, weight setting at model level
    model_config = model.get_config()
    recovered_model = keras.models.Model.from_config(model_config)
    if model.weights:
        weights = model.get_weights()
        recovered_model.set_weights(weights)
        output = recovered_model.predict(input_data)
        np.testing.assert_allclose(output, actual_output, rtol=1e-3, atol=1e-6)

    # test training mode (e.g. useful for dropout tests)
    # Rebuild the model to avoid the graph being reused between predict() and
    # See b/120160788 for more details. This should be mitigated after 2.0.
    if validate_training:
        model = keras.models.Model(x, layer(x))
        if _thread_local_data.run_eagerly is not None:
            model.compile(
                'rmsprop',
                'mse',
                weighted_metrics=['acc'],
                run_eagerly=should_run_eagerly())
        else:
            model.compile('rmsprop', 'mse', weighted_metrics=['acc'])
        model.train_on_batch(input_data, actual_output)

    # test as first layer in Sequential API
    layer_config = layer.get_config()
    layer_config['batch_input_shape'] = input_shape
    layer = layer.__class__.from_config(layer_config)

    # Test adapt, if data was passed.
    if adapt_data is not None:
        layer.adapt(adapt_data)

    model = keras.models.Sequential()
    model.add(layer)
    actual_output = model.predict(input_data)
    actual_output_shape = actual_output.shape
    for expected_dim, actual_dim in zip(computed_output_shape,
                                        actual_output_shape):
        if expected_dim is not None:
            if expected_dim != actual_dim:
                raise AssertionError(
                    'When testing layer %s **after deserialization**, '
                    'for input %s, found output_shape='
                    '%s but expected to find inferred shape %s.\nFull kwargs: %s' %
                    (layer_cls.__name__,
                     x,
                     actual_output_shape,
                     computed_output_shape,
                     kwargs))
    if expected_output is not None:
        np.testing.assert_allclose(actual_output, expected_output,
                                   rtol=1e-3, atol=1e-6)

    # test serialization, weight setting at model level
    model_config = model.get_config()
    recovered_model = keras.models.Sequential.from_config(model_config)
    if model.weights:
        weights = model.get_weights()
        recovered_model.set_weights(weights)
        output = recovered_model.predict(input_data)
        np.testing.assert_allclose(output, actual_output, rtol=1e-3, atol=1e-6)

    # for further checks in the caller function
    return actual_output



import tensorflow as tf

from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.ops.localized_attention_nd import localized_attention_1d, localized_attention_2d


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class LocalizedAttentionLayer1D(DynastesBaseLayer):

    def __init__(self,
                 kernel_size=3,
                 num_heads=1,
                 strides=1,
                 dilation_rate=1,
                 multiquery_attention=False,
                 padding='same',
                 preshaped_q=True, **kwargs):
        """
            Args:
                kernel_size: size of patches to perform localized attention within
                num_heads: number of attention heads
                strides: the strides of the patch window, strides 2 halves output
                dilation_rate: the dilation_rate of the patch window
                padding: one of 'same' or 'valid'
                preshaped_q: True if q matches strided and padded kv
                    ex: kv: [B, 4, C]
                        strides = 2
                        q must be [B,2,C]
        """
        super(LocalizedAttentionLayer1D, self).__init__(**kwargs)
        self.kernel_size = kernel_size
        self.num_heads = num_heads
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.padding = padding
        self.preshaped_q = preshaped_q
        self.multiquery_attention = multiquery_attention

    def call(self, q, k, v, training=None, mask=None):
        if type(q) == list:
            if len(q) == 3:
                q, k, v = q
            else:
                raise SyntaxError
        return localized_attention_1d(q=q, k=k, v=v,
                                      num_heads=self.num_heads,
                                      strides=self.strides,
                                      dilation_rate=self.dilation_rate,
                                      padding=self.padding,
                                      preshaped_q=self.preshaped_q,
                                      multiquery_attention=self.multiquery_attention), {}

    def get_config(self):
        config = {'kernel_size': self.kernel_size,
                  'num_heads': self.num_heads,
                  'strides': self.strides,
                  'dilation_rate': self.dilation_rate,
                  'padding': self.padding,
                  'preshaped_q': self.preshaped_q,
                  'multiquery_attention': self.multiquery_attention}
        base_config = super(LocalizedAttentionLayer1D, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class LocalizedAttentionLayer2D(DynastesBaseLayer):

    def __init__(self,
                 kernel_size=(3, 3),
                 num_heads=1,
                 strides=(1, 1),
                 dilation_rate=(1, 1),
                 padding='same',
                 preshaped_q=True,
                 multiquery_attention=False,
                 **kwargs):
        """
            Args:
                kernel_size: size of patches to perform localized attention within
                num_heads: number of attention heads
                strides: the strides of the patch window, strides 2 halves output
                dilation_rate: the dilation_rate of the patch window
                padding: one of 'same' or 'valid'
                preshaped_q: True if q matches strided and padded kv
                    ex: kv: [B, 4, 4, C]
                        strides = (2,2)
                        q must be [B,2,2,C]
        """
        super(LocalizedAttentionLayer2D, self).__init__(**kwargs)
        self.kernel_size = kernel_size
        self.num_heads = num_heads
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.padding = padding
        self.preshaped_q = preshaped_q
        self.multiquery_attention = multiquery_attention

    def call(self, q, k, v, training=None, mask=None):
        if type(q) == list:
            if len(q) == 3:
                q, k, v = q
            else:
                raise SyntaxError
        return localized_attention_2d(q=q, k=k, v=v,
                                      num_heads=self.num_heads,
                                      strides=self.strides,
                                      dilation_rate=self.dilation_rate,
                                      padding=self.padding,
                                      preshaped_q=self.preshaped_q,
                                      multiquery_attention=self.multiquery_attention)

    def get_config(self):
        config = {'kernel_size': self.kernel_size,
                  'num_heads': self.num_heads,
                  'strides': self.strides,
                  'dilation_rate': self.dilation_rate,
                  'padding': self.padding,
                  'preshaped_q': self.preshaped_q,
                  'multiquery_attention': self.multiquery_attention}
        base_config = super(LocalizedAttentionLayer2D, self).get_config()
        return {**base_config, **config}
